# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import textwrap
from typing import (Any, Dict, List, Set, Tuple, Type, Union, TYPE_CHECKING)

from ..core import exceptions
from ..utility import conversion
from ..utility.mask import round_up_div

if TYPE_CHECKING:
    from .execution_context import CommandExecutionContext
    from ..core.core_registers import CoreRegisterInfo

LOG = logging.getLogger(__name__)

## @brief Dict of command group names to a set of command classes.
ALL_COMMANDS: Dict[str, Set[Union["CommandBase", "ValueBase"]]] = {}

class CommandMeta(type):
    """@brief Metaclass for commands.

    Examines the `INFO` attribute of the command class and builds the @ref pyocd.commands.commands.ALL_COMMANDS
    "ALL_COMMANDS" table.
    """

    def __new__(mcs: Type, name: str, bases: Tuple[type, ...], objdict: Dict[str, Any]) -> "CommandMeta":
        # Create the new type.
        new_type = type.__new__(mcs, name, bases, objdict)

        # The Command base class won't have an INFO.
        if 'INFO' in objdict:
            info = objdict['INFO']

            # Validate the INFO dict.
            assert (('names' in info)
                    and ('group' in info)
                    and ('category' in info)
                    and ('help' in info)
                    and ((('nargs' in info) and ('usage' in info)) # Required for commands.
                        or ('access' in info))) # Required for values.

            # Add this command to our table of commands by group.
            ALL_COMMANDS.setdefault(info['group'], set()).add(new_type)
        return new_type

class CommandBase(metaclass=CommandMeta):
    """@brief Base class for a command.

    Each command class must have an `INFO` attribute with the following keys:
    - `names`: List of names for the info. The first element is the primary name.
    - `group`: Optional name for the command group. The group is meant to be a context group, not type of
        command. Most commands belong to the 'standard' group.
    - `category`: Functional category for the command.
    - `nargs`: The number of arguments the command accepts. Either the string "*", meaning any number of
        arguments, a single integer, or a sequence of integers.
    - `usage`: String with a description of the command's argument usage.
    - `help`: String for the short help. Typically should be no more than one sentence.
    - `extra_help`: Optional key for a string with more detailed help.
    """

    ## Default, empty info dict.
    INFO = {
            'names': [],
            'group': '',
            'category': '',
            'nargs': 0,
            'usage': "",
            'help': "",
            }

    def __init__(self, context: "CommandExecutionContext") -> None:
        """@brief Constructor."""
        self._context = context

    @property
    def context(self):
        """@brief The command execution context."""
        return self._context

    def check_arg_count(self, args: List[str]) -> None:
        """@brief Verify the number of command arguments."""
        nargs = self.INFO['nargs']
        if nargs == '*':
            pass
        elif nargs is None:
            if len(args) != 0:
                raise exceptions.CommandError("command does not accept arguments")
        elif isinstance(nargs, list):
            if (len(args) not in nargs):
                raise exceptions.CommandError("incorrect number of arguments")
        elif len(args) < nargs:
            raise exceptions.CommandError("too few arguments")
        elif len(args) > nargs:
            raise exceptions.CommandError("too many arguments")

    def parse(self, args: List[str]) -> None:
        """@brief Extract command arguments."""
        pass

    def execute(self) -> None:
        """@brief Perform the command."""
        raise NotImplementedError()

    def _format_core_register(self, info: "CoreRegisterInfo", value: int) -> str:
        hex_width = round_up_div(info.bitsize, 4) + 2 # add 2 for the "0x" prefix
        if info.is_double_float_register:
            value_str = "{f:g} ({i:#0{w}x})".format(f=conversion.u64_to_float64(value), i=value, w=hex_width)
        elif info.is_single_float_register:
            value_str = "{f:g} ({i:#0{w}x})".format(f=conversion.u32_to_float32(value), i=value, w=hex_width)
        elif info.gdb_type in ('data_ptr', 'code_ptr'):
            value_str = "{h:#0{w}x}".format(h=value, w=hex_width)
        else:
            value_str = "{h:#0{w}x} ({d:d})".format(h=value, w=hex_width, d=value)
        return value_str

    def _convert_value(self, arg: str) -> int:
        """@brief Convert an argument to a 32-bit integer.

        Handles the usual decimal, binary, and hex numbers with the appropriate prefix.
        Also recognizes register names and address dereferencing. Dereferencing using the
        ARM assembler syntax. To dereference, put the value in brackets, i.e. '[r0]' or
        '[0x1040]'. You can also use put an offset in the brackets after a comma, such as
        '[r3,8]'. The offset can be positive or negative, and any supported base.
        """
        try:
            offset = 0
            deref = (arg[0] == '[')
            if deref:
                if not self.context.selected_core:
                    raise exceptions.CommandError("cannot dereference when memory is not accessible")
                arg = arg[1:-1]
                offset = 0
                if ',' in arg:
                    arg, offset = arg.split(',')
                    arg = arg.strip()
                    offset = int(offset.strip(), base=0)

            value = None
            if (self.context.selected_core) and (arg.lower() in self.context.selected_core.core_registers.by_name):
                value = self.context.selected_core.read_core_register(arg.lower())
                self.context.writei("%s = 0x%08x", arg.lower(), value)
            else:
                subargs = arg.lower().split('.')
                if subargs[0] in self.context.peripherals and len(subargs) > 1:
                    p = self.context.peripherals[subargs[0]]
                    r = [x for x in p.registers if x.name.lower() == subargs[1]]
                    if len(r):
                        value = p.base_address + r[0].address_offset
                    else:
                        raise exceptions.CommandError("invalid register '%s' for %s" % (subargs[1], p.name))
                elif self.context.elf is not None:
                    sym = self.context.elf.symbol_decoder.get_symbol_for_name(arg)
                    if sym is not None:
                        value = sym.address

            if value is None:
                arg = arg.lower().replace('_', '')
                value = int(arg, base=0)

            if deref and (self.context.selected_ap is not None):
                value = conversion.byte_list_to_u32le_list(
                        self.context.selected_ap.read_memory_block8(value + offset, 4))[0]
                self.context.writei("[%s,%d] = 0x%08x", arg, offset, value)

            return value
        except ValueError as err:
            raise exceptions.CommandError("invalid argument '{}'".format(arg)) from None

    @classmethod
    def format_help(cls, context, max_width: int = 72) -> str:
        """@brief Return a string with the help text for this command."""
        text = "Usage: {cmd} {usage}\n".format(cmd=cls.INFO['names'][0], usage=cls.INFO['usage'])
        if len(cls.INFO['names']) > 1:
            text += "Aliases: {0}\n".format(", ".join(cls.INFO['names'][1:]))
        text += "\n" + textwrap.fill(cls.INFO['help'], width=max_width) + "\n"
        if 'extra_help' in cls.INFO:
            text += "\n" + textwrap.fill(cls.INFO['extra_help'], width=max_width) + "\n"
        return text

class ValueBase(CommandBase):
    """@brief Base class for value commands.

    Value commands are special commands representing a value that can be read and/or written. They are used
    through the `show` and `set` commands. A value command has an associated access mode of read-only,
    write-only, or read-write. The access mode sets which of the `show` and `set` commands may be used with
    the value.

    Each value class must have an `INFO` attribute with the following keys:
    - `names`: List of names for the value. The first element is the primary name.
    - `group`: Optional name for the command group. The group is meant to be a context group, not type of
        command. Most values belong to the 'standard' group.
    - `category`: Functional category for the value.
    - `access`: A string of either "r" or "rw" indicating whether the value is read-only, write-only, or
        read-write.
    - `help`: String for the short help. Typically should be no more than one sentence.
    - `extra_help`: Optional key for a string with more detailed help.
    """

    def display(self, args: List[str]) -> None:
        """@brief Output the value of the info."""
        raise NotImplementedError()

    def modify(self, args: List[str]) -> None:
        """@brief Change the info to a new value."""
        raise NotImplementedError()

    @classmethod
    def format_help(cls, context, max_width: int = 72) -> str:
        """@brief Return a string with the help text for this command."""
        first_name = cls.INFO['names'][0]
        text = "Usage: "
        did_print_on_usage_line = False
        if 'r' in cls.INFO['access']:
            usage = cls.INFO.get('show_usage', "")
            text += "show {cmd} {usage}\n".format(cmd=first_name, usage=usage)
            did_print_on_usage_line = True
        if 'w' in cls.INFO['access']:
            indent = "       " if did_print_on_usage_line else ""
            usage = cls.INFO.get('set_usage', "VALUE")
            text += "{indent}set {cmd} {usage}\n".format(indent=indent, cmd=first_name, usage=usage)
        if len(cls.INFO['names']) > 1:
            text += "Aliases: {0}\n".format(", ".join(cls.INFO['names'][1:]))
        text += "\n" + textwrap.fill(cls.INFO['help'], width=max_width) + "\n"
        if 'extra_help' in cls.INFO:
            text += "\n" + textwrap.fill(cls.INFO['extra_help'], width=max_width) + "\n"
        return text




# pyOCD debugger
# Copyright (c) 2021 mikisama
# Copyright (C) 2021 Ciro Cattuto <ciro.cattuto@gmail.com>
# Copyright (C) 2021 Simon D. Levy <simon.d.levy@gmail.com>
# Copyright (C) 2022 Johan Carlsson <johan.carlsson@teenage.engineering>
# Copyright (C) 2022 Samuel Dewan
# Copyright (C) 2022 Zhengji Li
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import logging
import sys
from time import sleep
import time
from typing import List

from pyocd.core.helpers import ConnectHelper
from pyocd.core.soc_target import SoCTarget
from pyocd.debug.rtt import RTTControlBlock, RTTUpChannel, RTTDownChannel
from pyocd.subcommands.base import SubcommandBase
from pyocd.utility.cmdline import convert_session_options, int_base_0
from pyocd.utility.kbhit import KBHit


LOG = logging.getLogger(__name__)


class RTTSubcommand(SubcommandBase):
    """@brief `pyocd rtt` subcommand."""

    NAMES = ["rtt"]
    HELP = "SEGGER RTT Viewer/Logger."

    @classmethod
    def get_args(cls) -> List[argparse.ArgumentParser]:
        """@brief Add this subcommand to the subparsers object."""

        rtt_parser = argparse.ArgumentParser(cls.HELP, add_help=False)

        rtt_options = rtt_parser.add_argument_group("rtt options")
        rtt_options.add_argument("-a", "--address", type=int_base_0, default=None,
                                 help="Start address of RTT control block search range.")
        rtt_options.add_argument("-s", "--size", type=int_base_0, default=None,
                                 help="Size of RTT control block search range.")
        rtt_options.add_argument("--up-channel-id", type=int, default=0,
                                 help="Up channel ID.")
        rtt_options.add_argument("--down-channel-id", type=int, default=0,
                                 help="Down channel ID.")
        rtt_options.add_argument("-d", "--log-file", type=str, default=None,
                                 help="Log file name. When specified, logging mode is enabled.")

        return [cls.CommonOptions.COMMON, cls.CommonOptions.CONNECT, rtt_parser]

    def invoke(self) -> int:

        session = None
        kb = None

        try:
            session = ConnectHelper.session_with_chosen_probe(
                project_dir=self._args.project_dir,
                config_file=self._args.config,
                user_script=self._args.script,
                no_config=self._args.no_config,
                pack=self._args.pack,
                unique_id=self._args.unique_id,
                target_override=self._args.target_override,
                frequency=self._args.frequency,
                blocking=(not self._args.no_wait),
                connect_mode=self._args.connect_mode,
                options=convert_session_options(self._args.options),
                option_defaults=self._modified_option_defaults(),
                )

            if session is None:
                LOG.error("No target device available")
                return 1

            with session:

                target: SoCTarget = session.board.target

                control_block = RTTControlBlock.from_target(target,
                            address = self._args.address,
                            size = self._args.size)
                control_block.start()

                if len(control_block.up_channels) < 1:
                    LOG.error("No up channels.")
                    return 1

                LOG.info(f"{len(control_block.up_channels)} up channels and "
                         f"{len(control_block.down_channels)} down channels found")

                up_chan: RTTUpChannel = control_block.up_channels[self._args.up_channel_id]
                up_name = up_chan.name if up_chan.name is not None else ""
                LOG.info(f"Reading from up channel {self._args.up_channel_id} (\"{up_name}\")")

                # some targets might need this here
                #target.reset_and_halt()

                target.resume()

                # set up terminal input
                kb = KBHit()

                if self._args.log_file is None:
                    if len(control_block.down_channels) < 1:
                        LOG.error("No down channels.")
                        return 1
                    down_chan: RTTDownChannel = control_block.down_channels[self._args.down_channel_id]
                    down_name = down_chan.name if down_chan.name is not None else ""
                    LOG.info(f"Writing to down channel {self._args.down_channel_id} (\"{down_name}\")")

                    self.viewer_loop(up_chan, down_chan, kb)
                else:
                    self.logger_loop(up_chan, kb)

        except KeyboardInterrupt:
            pass

        finally:
            if session:
                session.close()
            if kb:
                kb.set_normal_term()

        return 0

    def logger_loop(self, up_chan, kb):

        LOG.info("start logging ... Press any key to stop")
        total_size = 0
        block_size = 0
        last_time = time.time()

        with open(self._args.log_file, 'wb') as log_file:

            while True:
                # poll at most 1000 times per second to limit CPU use
                sleep(0.001)

                # read data from up buffer
                data = up_chan.read()
                log_file.write(data)

                s = len(data)
                block_size += s
                total_size += s
                diff = time.time() - last_time
                if diff > 1.0:
                    print(f"Transfer rate: {block_size / 1000:.1f} KByte/s; Bytes written: {total_size / 1000:.0f} KByte", end="\r")
                    block_size = 0
                    last_time = time.time()

                # try to fetch character
                if kb.kbhit():
                    break

    def viewer_loop(self, up_chan, down_chan, kb):
        # byte array to send via RTT
        cmd = bytes()

        while True:
            # poll at most 1000 times per second to limit CPU use
            sleep(0.001)

            # read data from up buffer 0 (target -> host) and write to
            # stdout
            up_data: bytes = up_chan.read()
            sys.stdout.buffer.write(up_data)
            sys.stdout.buffer.flush()

            # try to fetch character
            if kb.kbhit():
                c: str = kb.getch()

                if ord(c) == 27: # process ESC
                    break
                elif c.isprintable() or c == '\n':
                    print(c, end="", flush=True)

                # add char to buffer
                cmd += c.encode("utf-8")

            # write buffer to target
            if not cmd:
                continue

            # write cmd buffer to down buffer 0 (host -> target)
            bytes_out = down_chan.write(cmd)
            cmd = cmd[bytes_out:]



# -*- coding: utf-8 -*-
# pyOCD debugger
# Copyright (c) 2019-2020 Arm Limited
# Copyright (c) 2020 Men Shiyun
# Copyright (c) 2020 Fraunhofer-Gesellschaft zur Förderung der angewandten Forschung e.V.
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import (dataclass, field)
from xml.etree.ElementTree import (ElementTree, Element)
import zipfile
import logging
import io
import errno
from pathlib import Path
from typing import (Any, Callable, Dict, List, IO, Optional, Tuple, TypeVar, Set, Union)

from .flash_algo import PackFlashAlgo
from ...core import exceptions
from ...core.memory_map import (
    FlashRegion,
    MemoryMap,
    MemoryRange,
    MemoryRegion,
    MemoryType,
    MEMORY_TYPE_CLASS_MAP,
)
from ...coresight.ap import (
    APAddressBase,
    APv1Address,
    APv2Address,
)
from ...debug.sequences.sequences import (
    Block,
    DebugSequence,
    DebugSequenceNode,
    IfControl,
    WhileControl,
)

LOG = logging.getLogger(__name__)

class MalformedCmsisPackError(exceptions.TargetSupportError):
    """@brief Exception raised for errors parsing a CMSIS-Pack."""
    pass

@dataclass
class _DeviceInfo:
    """@brief Simple container class to hold XML elements describing a device."""
    element: Element
    families: List[str] = field(default_factory=list)
    processors: List[Element] = field(default_factory=list)
    memories: List[Element] = field(default_factory=list)
    algos: List[Element] = field(default_factory=list)
    debugs: List[Element] = field(default_factory=list)
    sequences: List[Element] = field(default_factory=list)
    debugvars: List[Element] = field(default_factory=list)
    debugports: List[Element] = field(default_factory=list)
    accessports: List[Element] = field(default_factory=list)

@dataclass
class ProcessorInfo:
    """@brief Descriptor for a processor defined in a DFP."""
    ## The Pname attribute, or Dcore if not Pname was provided.
    name: str = "unknown"
    ## PE unit number within an MPCore. For single cores this will be 0.
    unit: int = -1
    ## Total number of cores in an MPCore.
    total_units: int = 1
    ## Address of AP through which the PE can be accessed.
    ap_address: APAddressBase = APv1Address(-1)
    ## Base address of the PE's memory mapped debug registers. Not used and 0 for M-profile.
    address: int = 0
    ## SVD file path relative to the pack.
    svd_path: Optional[str] = None
    ## Default reset sequence name.
    default_reset_sequence: str = "ResetSystem"

def _get_part_number_from_element(element: Element) -> str:
    """@brief Extract the part number from a device or variant XML element."""
    assert element.tag in ("device", "variant")
    # Both device and variant may have 'Dname' according to the latest spec.
    if 'Dname' in element.attrib:
        return element.attrib['Dname']
    elif element.tag == "variant":
        return element.attrib['Dvariant']
    else:
        raise ValueError("element is neither device nor variant")

class CmsisPack:
    """@brief Wraps a CMSIS Device Family Pack.

    This class provides a top-level interface for extracting device information from CMSIS-Packs.
    After an instance is constructed, a list of the devices described within the pack is available
    from the `devices` property. Each item in the list is a CmsisPackDevice object.

    The XML element hierarchy that defines devices is as follows.
    ```
    family [-> subFamily] -> device [-> variant]
    ```

    Internally, this class is responsible for collecting the device-related XML elements from each
    of the levels of the hierarchy described above. It determines which elements belong to each
    defined device and passes those to CmsisPackDevice. It is then CmsisPackDevice that performs
    the parsing of each element type into pyOCD-compatible data.
    """
    def __init__(self, file_or_path: Union[str, zipfile.ZipFile, IO[bytes], Path]) -> None:
        """@brief Constructor.

        Opens the CMSIS-Pack and builds instances of CmsisPackDevice for all the devices
        and variants defined within the pack.

        @param self
        @param file_or_path The .pack file to open. These values are supported:
            - String that is the path to a .pack file (a Zip file).
            - String that is the path to the root directory of an expanded pack.
            - `ZipFile` object.
            - File-like object that is already opened.

        @exception MalformedCmsisPackError The pack is not a zip file, or the .pdsc file is missing
            from within the pack.
        """
        self._is_dir = False
        if isinstance(file_or_path, zipfile.ZipFile):
            self._pack_file = file_or_path
        else:
            # Check for an expanded pack as a directory.
            if isinstance(file_or_path, (str, Path)):
                path = Path(file_or_path).expanduser()
                file_or_path = str(path) # Update with expanded path.

                self._is_dir = path.is_dir()
                if self._is_dir:
                    self._dir_path = path

            if not self._is_dir:
                try:
                    self._pack_file = zipfile.ZipFile(file_or_path, 'r')
                except zipfile.BadZipFile as err:
                    raise MalformedCmsisPackError(f"Failed to open CMSIS-Pack '{file_or_path}': {err}") from err

        # Find the .pdsc file.
        if self._is_dir:
            for child_path in self._dir_path.iterdir():
                if child_path.suffix == '.pdsc':
                    self._pdsc_name = child_path.name
                    break
            else:
                raise MalformedCmsisPackError(f"CMSIS-Pack '{file_or_path}' is missing a .pdsc file")
        else:
            for name in self._pack_file.namelist():
                if name.endswith('.pdsc'):
                    self._pdsc_name = name
                    break
            else:
                raise MalformedCmsisPackError(f"CMSIS-Pack '{file_or_path}' is missing a .pdsc file")

        if self._is_dir:
            with (self._dir_path / self._pdsc_name).open('rb') as pdsc_file:
                self._pdsc = CmsisPackDescription(self, pdsc_file)
        else:
            with self._pack_file.open(self._pdsc_name) as pdsc_file:
                self._pdsc = CmsisPackDescription(self, pdsc_file)

    @property
    def filename(self) -> Optional[str]:
        """@brief Accessor for the filename or path of the .pack file."""
        return self._pack_file.filename

    @property
    def pdsc(self) -> "CmsisPackDescription":
        """@brief Accessor for the CmsisPackDescription instance for the pack's PDSC file."""
        return self._pdsc

    @property
    def devices(self) -> List["CmsisPackDevice"]:
        """@brief A list of CmsisPackDevice objects for every part number defined in the pack."""
        return self._pdsc.devices

    def get_file(self, filename: str) -> IO[bytes]:
        """@brief Return file-like object for a file within the pack.

        @param self
        @param filename Relative path within the pack. May use forward or back slashes.
        @return A BytesIO object is returned that contains all of the data from the file
            in the pack. This is done to isolate the returned file from how the pack was
            opened (due to particularities of the ZipFile implementation).
        """
        filename = filename.replace('\\', '/')

        # Some vendors place their pdsc in some subdirectories of the pack archive,
        # use relative directory to the pdsc file while reading other files.
        pdsc_base = self._pdsc_name.rsplit('/', 1)
        if len(pdsc_base) == 2:
            filename = f'{pdsc_base[0]}/{filename}'

        if self._is_dir:
            path = self._dir_path / filename
            return io.BytesIO(path.read_bytes())
        else:
            return io.BytesIO(self._pack_file.read(filename))

class CmsisPackDescription:
    """@brief Parser for the PDSC XML file describing a CMSIS-Pack.
    """

    def __init__(self, pack: CmsisPack, pdsc_file: IO[bytes]) -> None:
        """@brief Constructor.

        @param self This object.
        @param pack Reference to the CmsisPack instance.
        @param pdsc_file A file-like object for the .pdsc contained in _pack_.
        """
        self._pack = pack

        # Convert PDSC into an ElementTree.
        self._pdsc = ElementTree(file=pdsc_file)

        self._state_stack: List[_DeviceInfo] = []
        self._devices: List["CmsisPackDevice"] = []

        # Remember if we have already warned about overlapping memory regions
        # so we can limit these to one warning per DFP
        self._warned_overlapping_memory_regions = False

        # Extract devices.
        for family in self._pdsc.iter('family'):
            self._parse_devices(family)

    @property
    def pack_name(self) -> Optional[str]:
        """@brief Name of the CMSIS-Pack.
        @return Contents of the required <name> element, or None if missing.
        """
        return self._pdsc.findtext('name')

    @property
    def pack(self) -> CmsisPack:
        """@brief Reference to the containing CmsisPack object."""
        return self._pack

    @property
    def devices(self) -> List["CmsisPackDevice"]:
        """@brief A list of CmsisPackDevice objects for every part number defined in the pack."""
        return self._devices

    def _parse_devices(self, parent: Element) -> None:
        # Extract device description elements we care about.
        newState = _DeviceInfo(element=parent)
        children: List[Element] = []
        for elem in parent:
            if elem.tag == 'memory':
                newState.memories.append(elem)
            elif elem.tag == 'processor':
                newState.processors.append(elem)
            elif elem.tag == 'algorithm':
                newState.algos.append(elem)
            elif elem.tag == 'debug':
                newState.debugs.append(elem)
            elif elem.tag == 'sequences':
                newState.sequences += elem.findall('sequence')
            elif elem.tag == 'debugvars':
                newState.debugvars.append(elem)
            elif elem.tag == 'debugport':
                newState.debugports.append(elem)
            elif elem.tag in ('accessportV1', 'accessportV2'):
                newState.accessports.append(elem)
            # Save any elements that we will recurse into.
            elif elem.tag in ('subFamily', 'device', 'variant'):
                children.append(elem)

        # Push the new device description state onto the stack.
        self._state_stack.append(newState)

        # Create a device object if this element defines one.
        if parent.tag in ('device', 'variant'):
            # Build device info from elements applying to this device.
            deviceInfo = _DeviceInfo(element=parent,
                                        families=self._extract_families(),
                                        processors=self._extract_processors(),
                                        memories=self._extract_memories(),
                                        algos=self._extract_algos(),
                                        debugs=self._extract_debugs(),
                                        sequences=self._extract_sequences(),
                                        debugvars=self._extract_debugvars(),
                                        debugports=self._extract_debugports(),
                                        accessports=self._extract_accessports(),
                                        )

            # Support ._pack being None for testing.
            dev = CmsisPackDevice(self, deviceInfo, self._pack.get_file if self._pack else None)
            self._devices.append(dev)

        # Recursively process subelements.
        for elem in children:
            self._parse_devices(elem)

        self._state_stack.pop()

    def _extract_families(self) -> List[str]:
        """@brief Generate list of family names for a device."""
        families = []
        for state in self._state_stack:
            elem = state.element
            if elem.tag == 'family':
                families += [elem.attrib['Dvendor'], elem.attrib['Dfamily']]
            elif elem.tag == 'subFamily':
                families += [elem.attrib['DsubFamily']]
        return families

    ## Typevar used for _extract_items().
    _V = TypeVar('_V')

    def _extract_items(
                self,
                state_info_name: str,
                filter: Callable[[Dict[Any, _V], Element], None]
            ) -> List[_V]:
        """@brief Generic extractor utility.

        Iterates over saved elements for the specified device state info for each level of the
        device state stack, from outer to inner, calling the provided filter callback each
        iteration. A dictionary object is created and repeatedly passed to the filter callback, so
        state can be stored across calls to the filter.

        The general idea is that the filter callback extracts some identifying information from the
        element it is given and uses that as a key in the dictionary. When the filter is called for
        more deeply nested elements, those elements will override the any previously examined
        elements with the same identifier.

        @return All values from the dictionary.
        """
        map = {}
        for state in self._state_stack:
            for elem in getattr(state, state_info_name):
                try:
                    filter(map, elem)
                except (KeyError, ValueError) as err:
                    LOG.debug("error parsing CMSIS-Pack %s: %s", self.pack_name, err)
        return list(map.values())

    def _inherit_attributes(self, to_elem: Element, from_elem: Optional[Element]) -> Element:
        """@brief Add attributes missing from an elemnt but present in another.

        Copy to `to_elem` any attributes defined in `from_elem` but not defined, and therefore overridden,
        in `to_elem`.

        @param self
        @param to_elem The Element to which inherited attributes will be added.
        @param from_elem The Element from which attributes should be inherited. May be None, in which case
            `to_elem` is returned unmodified.
        @return The `to_elem` parameter is returned.
        """
        if from_elem is not None:
            inherited = {
                k: v
                for k, v in from_elem.attrib.items()
                if k not in to_elem.attrib
            }
            to_elem.attrib.update(inherited)
        return to_elem

    def _extract_processors(self) -> List[Element]:
        """@brief Extract processor elements.

        Attributes:
        - `Pname`: optional str for single-core devices
        - `Punits`: optional int, number of cores for MPCore cluster
        - `Dcore`: str, CPU type
        - plus a handful of others that specify processor options such as FPU or MVE
        """
        def filter(map: Dict, elem: Element) -> None:
            # Pname attribute is optional if there is only one CPU.
            pname = elem.attrib.get('Pname')
            map[pname] = self._inherit_attributes(elem, map.get(pname))

        return self._extract_items('processors', filter)

    def _extract_memories(self) -> List[Element]:
        """@brief Extract memory elements.

        The unique identifier is a bi-tuple of the memory's name, which is either the 'name' or 'id' attribute,
        in that order, plus the pname. If neither attribute exists, the region base and size are turned into
        a string.

        In addition to the name based filtering, memory regions are checked to prevent overlaps.

        Attributes:
        - `Pname`: optional str
        - `id`: optional str, deprecated in favour of `name`
        - `name`: optional str, overrides `id` if both are present
        - `access`: optional str
        - `start`: int
        - `size`: int
        - `default`: optional, if true indicates the region needs no special provisions for accessing,
            default false
        - `startup`: optional, if true use the region for boot code, default false
        - `init`: optional, deprecated, if true don't zeroise the memory, default false
        - `uninit`, optional, if true the memory should not be initialised, default false
        - `alias`: optional, another region's name
        """
        def get_start_and_size(elem: Element) -> Tuple[int, int]:
            try:
                start = int(elem.attrib['start'], base=0)
                size = int(elem.attrib['size'], base=0)
            except (KeyError, ValueError):
                LOG.warning("memory region missing address")
                raise
            return (start, size)

        def filter(map: Dict, elem: Element) -> None:
            # Inner memory regions are allowed to override outer memory regions.
            # If this is not done properly via name/id, we must make
            # sure not to report overlapping memory regions to gdb since it
            # will ignore those completely, see:
            # https://github.com/pyocd/pyOCD/issues/980
            start, size = get_start_and_size(elem)
            if 'name' in elem.attrib: # 'name' takes precedence over 'id'.
                name = elem.attrib['name']
            elif 'id' in elem.attrib:
                name = elem.attrib['id']
            else:
                # Neither option for memory name was specified, so use the address range.
                # Use the start and size for a name.
                name = "%08x:%08x" % (start, size)

            pname = elem.attrib.get('Pname', None)
            info = (name, pname)

            if info in map:
                del map[info]
            for k in list(map.keys()):
                prev_pname = k[1]
                # Previously, we would not check for overlaps if the pname was different. But because pyocd
                # currently only supports one memory map for the whole device, we have to ignore the pname for
                # now.
                prev_elem = map[k]
                prev_start, prev_size = get_start_and_size(prev_elem)
                # Overlap: start or end between previous start and previous end
                end = start + size - 1
                prev_end = prev_start + prev_size - 1
                if (prev_start <= start < prev_end) or (prev_start <= end < prev_end):
                    # Only report warnings for overlapping regions from the same processor. Allow regions for
                    # different processors to override each other, since we don't yet support maps for each
                    # processor.
                    if (pname == prev_pname) and not self._warned_overlapping_memory_regions:
                        filename = self.pack.filename if self.pack else "unknown"
                        LOG.warning("Overlapping memory regions in file %s (%s); deleting outer region. "
                                    "Further warnings will be suppressed for this file.",
                                    filename, _get_part_number_from_element(self._state_stack[-1].element))
                        self._warned_overlapping_memory_regions = True
                    del map[k]

            map[info] = elem

        return self._extract_items('memories', filter)

    def _extract_algos(self) -> List[Element]:
        """@brief Extract algorithm elements.

        The unique identifier is the algorithm's memory address range.

        Any algorithm elements with a 'style' attribuet not set to 'Keil' (case-insensitive) are
        skipped.

        Attributes:
        - `Pname`: optional str
        - `name`: str
        - `start`: int
        - `size`: int
        - `RAMstart`: optional int
        - `RAMsize`: optional int
        - `default`: optional bool
        - `style`: optional str
        """
        def filter(map: Dict, elem: Element) -> None:
            # We only support Keil FLM style flash algorithms (for now).
            if ('style' in elem.attrib) and (elem.attrib['style'].lower() != 'keil'):
                LOG.debug("%s DFP: skipping non-Keil flash algorithm", self.pack_name)
                return

            # Both start and size are required.
            start = int(elem.attrib['start'], base=0)
            size = int(elem.attrib['size'], base=0)
            memrange = (start, size)

            # An algo with the same range as an existing algo will override the previous.
            map[memrange] = elem

        return self._extract_items('algos', filter)

    def _extract_debugs(self) -> List[Element]:
        """@brief Extract debug elements.

        If the debug element does not have a 'Pname' element, its identifier is set to "*" to
        represent that it applies to all processors.

        Otherwise, the identifier is the element's 'Pname' attribute combined with 'Punit' if
        present. When 'Pname' is detected and a "*" key is in the map, the map is cleared before
        adding the current element.

        Attributes:
        - `__dp`: optional int
        - `__ap`: optional int
        - `__apid`: optional int
        - `address`: optional int
        - `svd`: optional str
        - `Pname`: optional str
        - `Punit`: optional int
        - `defaultResetSequence`: optional str

        Can have `<datapatch>` elements as children.
        """
        def filter(map: Dict, elem: Element) -> None:
            if 'Pname' in elem.attrib:
                name = elem.attrib['Pname']
                unit = elem.attrib.get('Punit', 0)
                name += str(unit)

                if '*' in map:
                    map.clear()

                map[name] = self._inherit_attributes(elem, map.get(name))
            else:
                # No processor name was provided, so this debug element applies to
                # all processors (well, there should only be one in this case).
                new_elem = self._inherit_attributes(elem, map.get('*'))
                map.clear()
                map['*'] = new_elem

        return self._extract_items('debugs', filter)

    def _extract_sequences(self) -> List[Element]:
        """@brief Extract debug sequence elements.

        The unique identifier is the sequence's 'name' attribute, combined with 'Pname' if present.

        Attributes:
        - `name`: str
        - `Pname`: optional str
        - `disable`: optional bool
        - `info`: optional str
        """
        def filter(map: Dict, elem: Element) -> None:
            if 'name' not in elem.attrib:
                LOG.debug("skipping unnamed debug sequence")
                return

            # Combine name and Pname.
            name = elem.attrib['name']
            pname = elem.attrib.get('Pname', None)

            map[(name, pname)] = elem

        return self._extract_items('sequences', filter)

    def _extract_debugvars(self) -> List[Element]:
        """@brief Extract debugvar elements.

        Works similar to _extract_debugs(), where only a single debugvar element is allowed unless
        the 'Pname' attribute appears in one of the elements.

        Attributes:
        - `configfile`: optional str
        - `version`: optional str
        - `Pname`: optional str
        """
        def filter(map: Dict, elem: Element) -> None:
            # No point in tracking an empty debugvars.
            if elem.text is None:
                return
            if 'Pname' in elem.attrib:
                name = elem.attrib['Pname']

                if '*' in map:
                    map.clear()
                map[name] = elem
            else:
                # No processor name was provided, so this debugvar element applies to
                # all processors.
                map.clear()
                map['*'] = elem

        return self._extract_items('debugvars', filter)

    def _extract_debugports(self) -> List[Element]:
        """@brief Extract debugport elements.

        Attributes:
        - `__dp`: int

        Children:
        - `<swd>`
        - `<jtag>`
        - `<cjtag>`
        """
        def filter(map: Dict, elem: Element) -> None:
            map[elem.attrib['__dp']] = elem

        return self._extract_items('debugports', filter)

    def _extract_accessports(self) -> List[Element]:
        """@brief Extract accessportV1 and accessportV2 elements.

        Attributes for `<accessportV1>`:
        - `__apid`: int
        - `__dp`: optional int
        - `index`: int

        Attributes for `<accessportV2>`:
        - `__apid`: int
        - `__dp`: optional int
        - `address`: int
        - `parent`: optional int
        """
        def filter(map: Dict, elem: Element) -> None:
            map[elem.attrib['__apid']] = elem

        return self._extract_items('accessports', filter)

def _get_bool_attribute(elem: Element, name: str, default: bool = False) -> bool:
    """@brief Extract an XML attribute with a boolean value.

    Supports "true"/"false" or "1"/"0" as the attribute values. Leading and trailing whitespace
    is stripped, and the comparison is case-insensitive.

    @param elem ElementTree.Element object.
    @param name String for the attribute name.
    @param default An optional default value if the attribute is missing. If not provided,
        the default is False.
    """
    if name not in elem.attrib:
        return default
    else:
        value = elem.attrib[name].strip().lower()
        if value in ("true", "1"):
            return True
        elif value in ("false", "0"):
            return False
        else:
            return default

class CmsisPackDevice:
    """@brief Wraps a device defined in a CMSIS Device Family Pack.

    Responsible for converting the XML elements that describe the device into objects
    usable by pyOCD. This includes the memory map and flash algorithms. All extraction of data
    into usuable data structures is done lazily, since a CmsisPackDevice instance will be
    created for every device in installed DFPs but only one will actually be used per session.

    An instance of this class can represent either a `<device>` or `<variant>` XML element from
    the PDSC.
    """

    def __init__(self, pdsc: CmsisPackDescription, device_info: _DeviceInfo,
            get_pack_file_cb: Optional[Callable[[str], IO[bytes]]]) -> None:
        """@brief Constructor.
        @param self
        @param pdsc The CmsisPackDescription object that contains this device.
        @param device_info A _DeviceInfo object with the XML elements that describe this device.
        @param get_pack_file_cb Callable taking a relative filename and returning an open bytes file. May
            raise IOError exceptions. If not supplied, then no flash algorithms or other files used by
            the device description are accessible (primarily for testing).
        """
        self._pdsc = pdsc
        self._info: _DeviceInfo = device_info
        self._get_pack_file_cb = get_pack_file_cb
        self._part: str = _get_part_number_from_element(device_info.element)
        self._regions: List[MemoryRegion] = []
        self._saw_startup: bool = False
        self._default_ram: Optional[MemoryRegion] = None
        self._memory_map: Optional[MemoryMap] = None
        self._processed_algos: Set[Element] = set() # Algo elements we've converted to regions.
        self._sequences: Set[DebugSequence] = set()
        self._debugvars: Optional[Block] = None
        self._valid_dps: List[int] = []
        self._apids: Dict[int, APAddressBase] = {}
        self._processors_map: Dict[str, ProcessorInfo] = {}
        self._processors_ap_map: Dict[APAddressBase, ProcessorInfo] = {}
        self._built_apid_map: bool = False

    def _build_memory_regions(self) -> None:
        """@brief Creates memory region instances for the device.

        For each `<memory>` element in the device info, a memory region object is created and
        added to the `_regions` attribute. IROM or non-writable memories are created as RomRegions
        by this method. They will be converted to FlashRegions by _build_flash_regions().
        """
        for elem in self._info.memories:
            try:
                # Get the region name, type, and access permissions.
                if 'name' in elem.attrib:
                    name = elem.attrib['name']
                    access = elem.attrib['access']

                    if ('p' in access):
                        type = MemoryType.DEVICE
                    elif ('w' in access):
                        type = MemoryType.RAM
                    else:
                        type = MemoryType.ROM
                elif 'id' in elem.attrib:
                    name = elem.attrib['id']

                    if 'RAM' in name:
                        access = 'rwx'
                        type = MemoryType.RAM
                    else:
                        access = 'rx'
                        type = MemoryType.ROM
                else:
                    continue

                # Both start and size are required attributes.
                start = int(elem.attrib['start'], base=0)
                size = int(elem.attrib['size'], base=0)

                is_default = _get_bool_attribute(elem, 'default')
                is_startup = _get_bool_attribute(elem, 'startup')
                if is_startup:
                    self._saw_startup = True

                attrs = {
                        'name': name,
                        'start': start,
                        'length': size,
                        'access': access,
                        'is_default': is_default,
                        'is_boot_memory': is_startup,
                        'is_testable': is_default,
                        'alias': elem.attrib.get('alias', None),
                    }

                # Look for matching flash algo.
                try:
                    # TODO multiple matching algos per region
                    algo_element = self._find_matching_algo(MemoryRange(attrs['start'],
                                                            length=attrs['length']))
                except KeyError:
                    # Must be a mask ROM or non-programmable flash.
                    algo_element = None

                # Convert the region to flash if we found a matching algorithm element.
                if (algo_element is not None) and self._set_flash_attributes(algo_element, attrs):
                    # Mark this algo as processed.
                    self._processed_algos.add(algo_element)

                    # Since this region has an algo, it's now flash.
                    type = MemoryType.FLASH

                    # If we don't have a boot memory yet, pick the first flash.
                    if not self._saw_startup:
                        attrs['is_boot_memory'] = True
                        self._saw_startup = True

                # Create the memory region and add to map.
                region = MEMORY_TYPE_CLASS_MAP[type](**attrs)
                self._regions.append(region)

                # Record the first default ram for use in flash algos.
                if (self._default_ram is None) and (type is MemoryType.RAM) and is_default:
                    self._default_ram = region
            except (KeyError, ValueError) as err:
                # Ignore errors.
                LOG.debug("ignoring error parsing memories for CMSIS-Pack devices %s: %s",
                    self.part_number, str(err))

        # Now create flash regions for any algos we didn't process.
        for algo in [a for a in self._info.algos if a not in self._processed_algos]:
            # Should this algo be loaded by default?
            is_default = _get_bool_attribute(algo, 'default')
            if not is_default:
                LOG.debug("%s DFP (%s): not loading non-default flash algorithm '%s'",
                    self.pack_description.pack_name, self.part_number, algo.attrib['name'])
                continue

            # Load flash algo from .FLM file so we can get its address range, etc.
            pack_algo = self._load_flash_algo(algo.attrib['name'])
            if pack_algo is None:
                LOG.warning(f"{self.pack_description.pack_name} DFP ({self.part_number}): "
                    f"failed to find or load flash algorithm '{algo.attrib['name']}'")
                continue

            ram_attrs = self._get_flash_ram_attributes(algo)

            # If we don't have a boot memory yet, pick the first flash.
            # TODO this should be refactored to use the flash region with lowest address.
            rgn_attrs: Dict[str, Any] = {}
            if not self._saw_startup:
                rgn_attrs['is_boot_memory'] = True
                self._saw_startup = True

            # Create the memory region.
            region = FlashRegion(
                        name=pack_algo.flash_info.name.decode(encoding='ascii'),
                        start=pack_algo.flash_start,
                        length=pack_algo.flash_size,
                        access='rx',
                        flm=pack_algo,
                        sector_size=0,
                        # Mark the flash memory as inaccessible at boot, just to be safe. There's
                        # no real way to be sure about this. The vendor should have create a
                        # <memory> element!
                        is_default=False,
                        # Similarly, disallow testing of this region since we're not sure. This will
                        # make it impossible to run functional tests on some devices without a user
                        # script to help out.
                        is_testable=False,
                        **rgn_attrs,
                        **ram_attrs,
                        )
            self._regions.append(region)

    def _get_flash_ram_attributes(self, algo_element: Element) -> Dict[str, int]:
        attrs: Dict[str, int] = {}
        if 'RAMstart' in algo_element.attrib:
            attrs['_RAMstart'] = int(algo_element.attrib['RAMstart'], base=0)
            if 'RAMsize' in algo_element.attrib:
                attrs['_RAMsize'] = int(algo_element.attrib['RAMsize'], base=0)
            else:
                LOG.warning(
                    f"{self.pack_description.pack_name} DFP ({self.part_number}): "
                    f"flash algorithm '{algo_element.attrib['name']}' has RAMstart but is "
                    "missing RAMsize")

        return attrs

    def _set_flash_attributes(self, algo_element: Element, attrs: dict) -> bool:
        # Load flash algo from .FLM file.
        pack_algo = self._load_flash_algo(algo_element.attrib['name'])
        if pack_algo is None:
            LOG.warning(f"{self.pack_description.pack_name} DFP ({self.part_number}): "
                f"failed to find or load flash algorithm '{algo_element.attrib['name']}'")
            return False

        attrs['flm'] = pack_algo

        # Save the algo element's RAM attributes in the region for later use in
        # CoreSightTarget.create_flash().
        ram_attrs = self._get_flash_ram_attributes(algo_element)
        attrs.update(ram_attrs)

        # Set sector size to a fixed value to prevent any possibility of infinite recursion due to
        # the default lambdas for sector_size and blocksize returning each other's value.
        attrs['sector_size'] = 0

        # We have at least a partially matching algo. Change type to flash.
        return True

    def _find_matching_algo(self, range: MemoryRange) -> Element:
        """@brief Searches for a flash algo overlapping the regions's address range.

        The algo and region's ranges just have to overlap. There are some DFPs that specify algos that
        don't fully cover the region range, and potentially vice versa. It is possible that one algo
        covers more than one region, but that is handled by the method calling this one (per region).
        """
        for algo in self._info.algos:
            try:
                # Both start and size are required attributes.
                algo_range = MemoryRange(start=self._get_int_attribute(algo, 'start'),
                                            length=self._get_int_attribute(algo, 'size'))
            except MalformedCmsisPackError:
                # Ignore this algorithm. A warning has already been logged.
                continue

            # Check if the region and the algo overlap.
            if range.intersects_range(range=algo_range):
                # Verify this is a valid algorithm specification.
                if 'name' not in algo.attrib:
                    LOG.debug(
                        f"{self.pack_description.pack_name} DFP ({self.part_number}): flash algorithm "
                        f"covering {algo_range.start:x}-{algo_range.end:x} missing required 'name' element")
                else:
                    return algo
        raise KeyError("no matching flash algorithm")

    def _load_flash_algo(self, filename: str) -> Optional[PackFlashAlgo]:
        """@brief Return the PackFlashAlgo instance for the given flash algo filename."""
        try:
            algo_data = self.get_file(filename)
            return PackFlashAlgo(algo_data)
        except FileNotFoundError:
            # Return default value.
            return None

    def get_file(self, filename: str) -> IO[bytes]:
        """@brief Return file-like object for a file within the containing pack.

        @param self
        @param filename Relative path within the pack. May use forward or back slashes.
        @return A BytesIO object is returned that contains all of the data from the file
            in the pack. This is done to isolate the returned file from how the pack was
            opened (due to particularities of the ZipFile implementation).

        @exception OSError A problem occurred opening the file.
        @exception FileNotFoundError In addition to the usual case of the file actually not being found,
            this exception is raised if no `get_pack_file_cb` was passed to the constructor.
        """
        if self._get_pack_file_cb:
            return self._get_pack_file_cb(filename)
        else:
            raise FileNotFoundError(errno.ENOENT, "No such file or directory", filename)

    def _build_sequences(self):
        """@brief Convert 'sequence' elements into DebugSequenceNode objects."""
        assert not len(self._sequences)
        for elem in self._info.sequences:
            # Extract sequence name.
            try:
                name = elem.attrib['name']
            except KeyError:
                LOG.warning("invalid debug sequence; missing name")
                continue

            try:
                # Extract optional sequence attributes.
                is_enabled = not _get_bool_attribute(elem, 'disable', False)
                pname = elem.attrib.get('Pname', None)
                info = elem.attrib.get('info', "")

                # Create the top level sequence node.
                sequence = DebugSequence(name, is_enabled, pname, info)

                # Start processing subelements in the sequence.
                for child in elem:
                    self._build_one_sequence_node(sequence, child)

                # Save the complete sequence object.
                self._sequences.add(sequence)
            except KeyError:
                LOG.debug("invalid debug sequence")

    def _build_one_sequence_node(self, parent: DebugSequenceNode, elem: Element) -> None:
        """@brief Convert one 'sequence' element into a DebugSequenceNode object."""
        # Grab optional info text.
        info = elem.attrib.get('info', "")

        # Generate a sequence node based on the element type.
        if elem.tag == 'block':
            # No reason to create a Block if there is no code within it.
            if elem.text is not None:
                # Create and attach a block node. No subelements are allowed.
                is_atomic = _get_bool_attribute(elem, 'atomic', False)
                node = Block(elem.text, is_atomic, info)
                parent.add_child(node)
        elif elem.tag == 'control':
            # The attribute name determines the control node's function.
            if 'if' in elem.attrib:
                node = IfControl(elem.attrib['if'], info)
            elif 'while' in elem.attrib:
                node = WhileControl(elem.attrib['while'], info, int(elem.attrib.get('timeout', "0")))
            else:
                root_node = parent.find_root()
                assert isinstance(root_node, DebugSequence)
                LOG.warning("invalid 'control' node in debug sequence '%s'", root_node.name)
                return

            # Attach the new node.
            parent.add_child(node)

            # Process control node subelements recursively.
            for child in elem:
                self._build_one_sequence_node(node, child)
        else:
            root_node = parent.find_root()
            assert isinstance(root_node, DebugSequence)
            LOG.warning("unexpected XML element '%s' in debug sequence '%s'", elem.tag, root_node.name)

    @property
    def pack_description(self) -> CmsisPackDescription:
        """@brief The CmsisPackDescription object that defines this device."""
        return self._pdsc

    @property
    def part_number(self) -> str:
        """@brief Part number for this device.

        This value comes from either the `Dname` or `Dvariant` attribute, depending on whether the
        device was created from a `<device>` or `<variant>` element.
        """
        return self._part

    @property
    def vendor(self) -> str:
        """@brief Vendor or manufacturer name."""
        return self._info.families[0].split(':')[0]

    @property
    def families(self) -> List[str]:
        """@brief List of families the device belongs to, ordered most generic to least."""
        return [f for f in self._info.families[1:]]

    @property
    def memory_map(self) -> MemoryMap:
        """@brief MemoryMap object."""
        # Lazily construct the memory map.
        if self._memory_map is None:
            self._build_memory_regions()

            # Warn if there was no boot memory.
            if not self._saw_startup:
                LOG.warning("CMSIS-Pack device %s has no identifiable boot memory", self.part_number)

            self._memory_map = MemoryMap(self._regions)

        return self._memory_map

    @property
    def svd(self) -> Optional[IO[bytes]]:
        """@brief File-like object for the device's SVD file.
        @todo Support multiple cores.
        """
        try:
            svdPath = self._info.debugs[0].attrib['svd']
            return self.get_file(svdPath)
        except (KeyError, IndexError):
            return None

    @property
    def sequences(self) -> Set[DebugSequence]:
        """@brief Dictionary of defined sequence elements.

        The dictionary key is the sequence name, value is a DebugSequence instance.
        """
        # Lazily construct.
        if (not len(self._sequences)) and len(self._info.sequences):
            self._build_sequences()
        return self._sequences

    @property
    def debug_vars_sequence(self) -> Optional[Block]:
        """@brief A debug sequence Block for the 'debugvars' element."""
        # Lazily construct.
        if (self._debugvars is None) and len(self._info.debugvars):
            elem = self._info.debugvars[0]
            assert elem.text is not None # Ensured by CmsisPackDescription._extract_debugvars.
            self._debugvars = Block(elem.text, info="debugvars")
        return self._debugvars

    @property
    def valid_dps(self) -> List[int]:
        """@brief List of valid DP indices.

        If the device has only one DP, its number will be 0.
        """
        if not self._valid_dps:
            self._build_valid_dps()
        return self._valid_dps

    @property
    def uses_apid(self) -> bool:
        """@brief Whether use of __apid values is required.

        This is based on whether the DFP defined accessportV1/2 elements.
        """
        return len(self.apid_map) > 0

    @property
    def apid_map(self) -> Dict[int, APAddressBase]:
        """@brief Map of AP unique IDs to AP address objects.

        These AP unique IDs (__apid) are defined by the DFP author and are valid only within the scope
        of the DFP. They are referenced by the __apid special variable used in debug sequences.
        """
        # A bool attr is used here instead of the usual empty test for the other lazy properties
        # because the apid map can be empty, in which case apids are not used by this device.
        if not self._built_apid_map:
            self._build_aps_map()
        return self._apids

    @property
    def processors_map(self) -> Dict[str, ProcessorInfo]:
        """@brief Map of processor names to processor info objects.

        If the device has only one PE, it may not have a defined Pname. In that case, the Dcore
        attribute will be used in place of the Pname.
        """
        if not self._processors_map:
            self._build_aps_map()
        return self._processors_map

    @property
    def processors_ap_map(self) -> Dict[APAddressBase, ProcessorInfo]:
        """@brief Map from AP address to processor info objects"""
        if not self._processors_ap_map:
            self._processors_ap_map = {
                proc.ap_address: proc
                for proc in self.processors_map.values()
            }
        return self._processors_ap_map

    def _get_int_attribute(
                self,
                elem: Element,
                name: str,
                default: Optional[int] = None
            ) -> int:
        """@brief Retrieve a DFP XML element's attribute value as an integer.
        @exception MalformedCmsisPackError Raised if the attribute is missing but is required (no default
            was provided), or cannot be converted to an integer.
        """
        if name not in elem.attrib:
            if default is None:
                raise MalformedCmsisPackError(
                        f"{self.pack_description.pack_name} DFP ({self.part_number}): <{elem.tag}> missing "
                        f"required '{name}' attribute")
            else:
                return default

        try:
            value = int(elem.attrib[name], base=0)
            return value
        except ValueError:
            raise MalformedCmsisPackError(
                    f"{self.pack_description.pack_name} DFP ({self.part_number}): <{elem.tag}> '{name}' "
                    f"attribute is invalid ('{elem.attrib[name]}')")

    def _build_valid_dps(self) -> None:
        """@brief Extract list of DP indices defined for the device."""
        # Build list of defined DPs.
        for debugport in self._info.debugports:
            try:
                self._valid_dps.append(self._get_int_attribute(debugport, '__dp'))
            except MalformedCmsisPackError as err:
                LOG.warning("%s", err)

        # If no (valid) <debugport> elements are defined, just use a default of 0.
        if len(self._valid_dps) == 0:
            self._valid_dps.append(0)

    def _handle_accessports(self) -> None:
        """@brief Process accessportV1 and accessportV2 elements.

        The ._apids attribute is filled in with information from accessportVx elements.
        """
        # Process accessportV1 and accessportV2 elements.
        for accessport in self._info.accessports:
            try:
                # Get the __dp attribute, with a default of 0.
                ap_dp = self._get_int_attribute(accessport, '__dp', 0)

                # Validate __dp, but be forgiving and only log a warning.
                if ap_dp not in self.valid_dps:
                    LOG.warning(
                            f"{self.pack_description.pack_name} DFP ({self.part_number}): <{accessport.tag}> "
                            f"'__dp' attribute is invalid ({ap_dp})")

                # APv1
                if accessport.tag == 'accessportV1':
                    index = self._get_int_attribute(accessport, 'index')
                    ap_address = APv1Address(index, ap_dp)
                # APv2
                elif accessport.tag == 'accessportV2':
                    address = self._get_int_attribute(accessport, 'address')
                    ap_address = APv2Address(address, ap_dp)
                else:
                    raise exceptions.InternalError(
                            f"unexpected element <{accessport.tag}> in access ports list")

                # Save this AP address and the specified __apid.
                apid = self._get_int_attribute(accessport, '__apid')
                self._apids[apid] = ap_address
            except MalformedCmsisPackError as err:
                LOG.warning("%s", err)

    def _handle_processors(self) -> None:
        """@brief Extract processor definitions."""
        for proc in self._info.processors:
            try:
                # Get the processor name.
                if 'Pname' in proc.attrib:
                    pname = proc.attrib['Pname']
                elif 'Dcore' in proc.attrib:
                    pname = proc.attrib['Dcore']
                else:
                    raise MalformedCmsisPackError(
                            f"{self.pack_description.pack_name} DFP ({self.part_number}): <{proc.tag}> is "
                            "missing 'Dcore' attribute")

                # Get optional number of processor units.
                punits = self._get_int_attribute(proc, 'Punits', 1)

                # Check for an existing processor with the same name.
                if pname in self._processors_map:
                    LOG.warning(
                            f"{self.pack_description.pack_name} DFP ({self.part_number}): <processor> "
                            f"element has duplicate name '{pname}'")
                    continue

                # Add the processor, with temp AP and base address.
                self._processors_map[pname] = ProcessorInfo(name=pname, total_units=punits)

            except MalformedCmsisPackError as err:
                LOG.warning("%s", err)

        # At least one processor must have been defined.
        if len(self._processors_map) == 0:
            LOG.warning(
                    f"{self.pack_description.pack_name} DFP ({self.part_number}): no <processor> "
                    "elements were found")

            # Add dummy processor.
            self._processors_map["unknown"] = ProcessorInfo(name="unknown")

    def _build_aps_map(self) -> None:
        """@brief Extract map of __apids and Pname to AP.

        Process the <debugport>, <accessportVx>, and <debug> elements to extract maps of __apid values
        to AP addresses. Also extracts a map of processor Pname values to AP addresses.
        """
        # Go ahead and set this flag so we don't try to rebuild the map in case there's an error.
        self._built_apid_map = True

        self._handle_accessports()
        self._handle_processors()
        assert len(self._processors_map)

        for debug in self._info.debugs:
            try:
                # Get the Pname attribute. If there isn't one, just use the first (there should only be
                # one) processor's name.
                if 'Pname' in debug.attrib:
                    pname = debug.attrib['Pname']
                else:
                    pname = list(self._processors_map.keys())[0]

                if pname not in self._processors_map:
                    raise MalformedCmsisPackError(
                            f"{self.pack_description.pack_name} DFP ({self.part_number}): <{debug.tag}> "
                            f"references undefined processor name ('{pname}')")

                # Check for __apid attribute.
                if '__apid' in debug.attrib:
                    apid = self._get_int_attribute(debug, '__apid')
                    if apid not in self._apids:
                        raise MalformedCmsisPackError(
                                f"{self.pack_description.pack_name} DFP ({self.part_number}): <{debug.tag}> "
                                f"references undefined '__apid' ({apid})")

                    ap_address = self._apids[apid]
                # Fall back to __ap address with optional __dp.
                elif '__ap' in debug.attrib:
                    # Get and validate optional __dp.
                    dp_index = self._get_int_attribute(debug, '__dp', 0)
                    if dp_index not in self.valid_dps:
                        raise MalformedCmsisPackError(
                                f"{self.pack_description.pack_name} DFP ({self.part_number}): <{debug.tag}> "
                                f"'__dp' attribute is invalid ({dp_index})")

                    ap_index = self._get_int_attribute(debug, '__ap')
                    ap_address = APv1Address(ap_index, dp_index)
                # Otherwise define a default AP #0.
                else:
                    ap_address = APv1Address(0)

                # Update address info for this processor.
                try:
                    proc = self._processors_map[pname]
                except KeyError:
                    raise MalformedCmsisPackError(
                            f"{self.pack_description.pack_name} DFP ({self.part_number}): <{debug.tag}> "
                            f"'Pname' attribute is invalid ({pname})")
                proc.unit = self._get_int_attribute(debug, 'Punit', 0)
                proc.ap_address = ap_address
                proc.address = self._get_int_attribute(debug, 'address', 0)
                proc.svd_path = debug.attrib.get('svd')
                if 'defaultResetSequence' in debug.attrib:
                    # Still need to validate the specified default reset sequence after this. If not
                    # set, the default value is 'ResetSystem'.
                    proc.default_reset_sequence = debug.attrib['defaultResetSequence']
            except MalformedCmsisPackError as err:
                LOG.warning("%s", err)

    def __repr__(self):
        return "<%s@%x %s>" % (self.__class__.__name__, id(self), self.part_number)




# pyOCD debugger
# Copyright (c) 2022 Yuntu Microelectronics
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)

FLASH_ALGO = {
    'load_address' : 0x20000000,

    # Flash algorithm as a hex string
    'instructions': [
    0xe7fdbe00,
    0x48414601, 0x4a406840, 0x29006050, 0x483fd005, 0x60104a3f, 0x4a3cb2c8, 0xbf006090, 0x6840483a,
    0x40102280, 0xd0f92800, 0x68404837, 0x4010220c, 0xd0012800, 0x47702001, 0xe7fc2000, 0x4604b570,
    0x4616460d, 0x69004833, 0x43082101, 0x61084931, 0x49324831, 0x48326008, 0x46086008, 0x08406840,
    0x60480040, 0x492b2003, 0xbf006008, 0x68804829, 0x0f800780, 0xd1f92803, 0x49262001, 0xbf006048,
    0x68c04824, 0xd1fb2801, 0x4822bf00, 0x21046880, 0x28044008, 0x2000d1f9, 0x6008491e, 0x491a4821,
    0x20006008, 0xffacf7ff, 0x4601bd70, 0x47702000, 0x201eb510, 0xffa4f7ff, 0xb510bd10, 0x481a4604,
    0x20106020, 0xff9cf7ff, 0xb5f7bd10, 0x4606b082, 0x9c04460f, 0x20084635, 0x97009001, 0xcc01bf00,
    0xcc01c501, 0x2002c501, 0xff8af7ff, 0xd0022800, 0xb0052001, 0x9901bdf0, 0x1a409800, 0x98009000,
    0xdcec2800, 0xe7f42000, 0x40010000, 0xfd9573f5, 0x40010200, 0x4007c000, 0x0000b631, 0x4006a000,
    0x0000c278, 0x00308300, 0x12345678, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000041,
    'pc_unInit': 0x200000af,
    'pc_program_page': 0x200000cf,
    'pc_erase_sector': 0x200000bf,
    'pc_eraseAll': 0x200000b5,

    'static_base' : 0x20000000 + 0x00000004 + 0x0000012c,
    'begin_stack' : 0x20002150,
    'end_stack' : 0x20001150,
    'begin_data' : 0x20000000 + 0x1000,
    'page_size' : 0x8,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    # Enable double buffering
    'page_buffers' : [
        0x20000140,
        0x20000148
    ],
    'min_program_length' : 0x8,

    # Relative region addresses and sizes
    'ro_start': 0x4,
    'ro_size': 0x12c,
    'rw_start': 0x130,
    'rw_size': 0x4,
    'zi_start': 0x134,
    'zi_size': 0x0,

    # Flash information
    'flash_start': 0x0,
    'flash_size': 0x80000,
    'sector_sizes': (
        (0x0, 0x400),
    )
}
class YTM32B1MD1(CoreSightTarget):

    VENDOR = "Yuntu Microelectronics"
    
    MEMORY_MAP = MemoryMap(
        FlashRegion(    start=0x00000000,  length=0x80000,      blocksize=0x400, is_boot_memory=True, algo=FLASH_ALGO),
        RamRegion(      start=0x20000000,  length=0x2000)
        )

    def __init__(self, session):
        super().__init__(session, self.MEMORY_MAP)



# pyOCD debugger
# Copyright (c) 2021 Huada Semiconductor Corporation
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...coresight.coresight_target import CoreSightTarget
from ...core.memory_map import (FlashRegion, RamRegion, MemoryMap)
from ...debug.svd.loader import SVDFile


#DEBUG_ACTIVE
DEBUG_ACTIVE = 0x40002038
DEBUG_ACTIVE_VAL = 0x00000FFF

FLASH_ALGO = {
    'load_address' : 0x20000000,
    # Flash algorithm as a hex string
    'instructions': [
    0xE00ABE00, 0x062D780D, 0x24084068, 0xD3000040, 0x1E644058, 0x1C49D1FA, 0x2A001E52, 0x4770D1F2,
    0x4770ba40, 0x4770bac0, 0x4d6db570, 0x06006828, 0x486c0f44, 0x68e98900, 0x0d400540, 0x02c90ac9,
    0x60e94301, 0xd0262c00, 0xf8bcf000, 0x21016828, 0x60284308, 0x050068e8, 0xf000d5fc, 0x6828f8b3,
    0x438821e0, 0xf0006028, 0x2c01f8ad, 0x2c02d006, 0x2c03d007, 0x2c04d008, 0xe008d10d, 0x21026828,
    0x6828e007, 0xe0042104, 0x21086828, 0x6828e001, 0x43882110, 0xf0006028, 0x6828f895, 0x02092107,
    0x60284388, 0xf88ef000, 0x21036828, 0x438802c9, 0xf0006028, 0x4c4cf879, 0x60202020, 0xf874f000,
    0x60602017, 0xf870f000, 0x60a0201b, 0xf86cf000, 0x60e04846, 0xf868f000, 0x61204845, 0xf864f000,
    0x61602018, 0xf860f000, 0x61a020f0, 0xf85cf000, 0x00c0207d, 0x200061e0, 0x2000bd70, 0xb5704770,
    0x6a204c39, 0xd4fc06c0, 0xf84ef000, 0x43c02000, 0xf0006320, 0x6a20f849, 0x43082103, 0x25006220,
    0x6a20602d, 0xd4fc06c0, 0xf83ef000, 0x20006325, 0xb570bd70, 0x46054c2c, 0x06c86a21, 0xf000d4fc,
    0x2000f833, 0x632043c0, 0xf82ef000, 0x08806a20, 0x1c800080, 0x26006220, 0x6a20602e, 0xd4fc06c0,
    0xf822f000, 0x20006326, 0xb5f7bd70, 0x460f4616, 0xf0002400, 0x4d1cf819, 0x63281e60, 0xf814f000,
    0x08806a28, 0x1c400080, 0xe0066228, 0x5d319800, 0x6a285501, 0xd4fc06c0, 0x42bc1c64, 0xf000d3f6,
    0x2000f803, 0xbdfe6328, 0x6ac1480f, 0x0c094a11, 0x18890409, 0x6ac162c1, 0x0c094a0f, 0x18890409,
    0x477062c1, 0x68814806, 0x0c094a0a, 0x18890409, 0x68816081, 0x0c094a08, 0x18890409, 0x47706081,
    0x40002000, 0x00100c00, 0x40020000, 0x00004650, 0x000222e0, 0x00005a5a, 0x0000a5a5, 0x00000000
    ],

    # Relative function addresses
    'pc_init': 0x20000029,
    'pc_unInit': 0x200000fb,
    'pc_program_page': 0x2000016b,
    'pc_erase_sector': 0x20000133,
    'pc_eraseAll': 0x200000ff,

    'static_base' : 0x20000000 + 0x00000020 + 0x000001dc,
    'begin_stack' : 0x20000400,
    'begin_data' : 0x20000000 + 0x800,
    'page_size' : 0x200,
    'analyzer_supported' : False,
    'analyzer_address' : 0x00000000,
    'page_buffers' : [0x20000800, 0x20001000],   # Enable double buffering
    'min_program_length' : 0x200,
  }


class HC32L136(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x10000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x2000)
        )

    def __init__(self, session):
        super(HC32L136, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L136.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32L130(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x10000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x2000)
        )

    def __init__(self, session):
        super(HC32L130, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32L130.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)


class HC32F030(CoreSightTarget):

    VENDOR = "HDSC"

    MEMORY_MAP = MemoryMap(
        FlashRegion( start=0x00000000, length=0x10000, sector_size=0x200,
                        is_boot_memory=True,
                        algo=FLASH_ALGO),
        RamRegion(   start=0x20000000, length=0x2000)
        )

    def __init__(self, session):
        super(HC32F030, self).__init__(session, self.MEMORY_MAP)
        self._svd_location = SVDFile.from_builtin("HC32F030.svd")

    def post_connect_hook(self):
        self.write32(DEBUG_ACTIVE, DEBUG_ACTIVE_VAL)



# pyOCD debugger
# Copyright (c) 2022 Samuel Dewan
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
import selectors
import socket
from typing import Optional, Sequence

from ..core.soc_target import SoCTarget
from ..core import exceptions
from ..debug.rtt import RTTControlBlock, RTTUpChannel, RTTDownChannel


class RTTChanWorker(ABC):
    """@brief Source and sink for data to be transferred over RTT. """

    @abstractmethod
    def write_up_data(self, data: bytes) -> int:
        """@brief Write data that has been received from an up channel to the
                  correct destination.

        @param data The data to be written.
        @return The number of bytes that were successfully written.
        """
        pass

    @abstractmethod
    def get_down_data(self) -> bytes:
        """@brief Get data that should be written to a down channel if there is
                  any.

        @return Data to be written to down channel.
        """
        pass

    @abstractmethod
    def close(self):
        """@brief Cleanup channel worker and close any file descriptors."""
        pass


class RTTChanTCPWorker(RTTChanWorker):
    """@brief Implementation of channel worker that forwards RTT data via a TCP
              socket. """

    port: int

    def __init__(self, port: int, listen: bool = True):
        """
        @param port The port to connect to or to listen for connects on.
        @param listen If true a server will be started to accept one connection
                      at a time on the given port. If false a connection will be
                      made as a TCP client to a server running on the given
                      port on localhost.
        """
        if listen:
            self.server = socket.socket()
            self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self.server.bind(('localhost', port))
            self.server.listen(1)
            self.server.setblocking(False)
            self.client = None
        else:
            self.server = None
            self.client = socket.create_connection(('localhost', port), timeout = 1.0)
            self.client.setblocking(False)

        self.port = port

    def _check_for_new_client(self):
        if self.server is None:
            return

        sel = selectors.DefaultSelector()
        sel.register(self.server, selectors.EVENT_READ, None)
        events = sel.select(timeout = 0)
        for key, _ in events:
            if key.fileobj == self.server:
                self.client, _ = self.server.accept()
                self.client.setblocking(False)

    def write_up_data(self, data: bytes):
        if self.client is None:
            self._check_for_new_client()
            if self.client is None:
                return 0

        return self.client.send(data)

    def get_down_data(self):
        if self.client is None:
            self._check_for_new_client()
            if self.client is None:
                return b''

        sel = selectors.DefaultSelector()
        sel.register(self.client, selectors.EVENT_READ, None)
        events = sel.select(timeout = 0)
        for key, _ in events:
            if key.fileobj == self.client:
                data = self.client.recv(4096)
                if not data:
                    # client socket closed at other end
                    self.client.close()
                    self.client = None
                return data

        return bytes()

    def close(self):
        if self.server is not None:
            self.server.close()
        if self.client is not None:
            self.client.close()

class RTTChanFileWorker(RTTChanWorker):
    """@brief Implementation of channel worker that write data from RTT channel
              to a file and optionally reads data from a file into an RTT
              channel. """


class RTTServer:
    """@brief Keeps track of polling for multiple active RTT channels and the
              sources and sinks of data for each channel. """
    control_block: RTTControlBlock
    workers: Optional[Sequence[Optional[RTTChanWorker]]]
    up_buffers: Optional[Sequence[bytes]]
    down_buffers: Optional[Sequence[bytes]]

    def __init__(self, target: SoCTarget, address: int, size: int,
                 control_block_id: bytes):
        """
        @param target The target with which RTT communication is desired.
        @param address Base address for control block search range.
        @param size Control block search range. If 0 the control block will be
                    expected to be located at the provided address.
        @param control_block_id The control block ID string to search for. Must
                                be at most 16 bytes long.  Will be padded with
                                zeroes if less than 16 bytes.
        """
        self.control_block = RTTControlBlock.from_target(target, address = address,
                                    size = size, control_block_id = control_block_id)

        self.workers = None
        self.up_buffers = None
        self.down_buffers = None

    def poll(self):
        """@brief Reads from and writes to active RTT channels. """
        if not self.running:
            # not yet started
            return

        for i, worker in enumerate(self.workers):
            if worker is None:
                continue

            # Read from up channel
            try:
                up_chan: RTTUpChannel = self.control_block.up_channels[i]
            except IndexError:
                pass
            else:
                self.up_buffers[i] += up_chan.read()

            # Write to worker
            bytes_written = worker.write_up_data(self.up_buffers[i])
            self.up_buffers[i] = self.up_buffers[i][bytes_written:]

            # Read from worker
            self.down_buffers[i] += worker.get_down_data()

            # Write data to down channel
            try:
                down_chan: RTTDownChannel = self.control_block.down_channels[i]
            except IndexError:
                pass
            else:
                bytes_out: int = down_chan.write(self.down_buffers[i])
                self.down_buffers[i] = self.down_buffers[i][bytes_out:]

    def start(self):
        """@brief Find and parse RTT control block. """
        self.control_block.start()

        num_up_chans: int = len(self.control_block.up_channels)
        num_down_chans: int = len(self.control_block.down_channels)
        num_chans: int = max(num_up_chans, num_down_chans)

        self.workers = [None] * num_chans
        self.up_buffers = [bytes()] * num_up_chans
        self.down_buffers = [bytes()] * num_down_chans

    def stop(self):
        """@brief Close all RTT workers. """
        if not self.running:
            return

        for i, worker in enumerate(self.workers):
            if worker is not None:
                worker.close()

        self.workers = None
        self.up_buffers = None
        self.down_buffers = None

    @property
    def running(self):
        """@brief True if RTT is started. """
        return self.workers is not None

    def add_server(self, port: int, channel: int):
        """@brief Start a new TCP server to communicate with a given RTT channel.

        @param port The port on which the server should listen for new connections.
        @param channel The RTT channel which should be exposed over TCP.
        """
        if not self.running:
            raise exceptions.RTTError("RTT is not yet started")
        elif self.workers[channel] is not None:
            raise exceptions.RTTError(f"RTT is already started for channel {channel}")

        self.workers[channel] = RTTChanTCPWorker(port, listen = True)

    def stop_server(self, port: int):
        """@brief Stop a TCP server.

        @param port The port of the server to be stopped.
        """

        if not self.running:
            return

        for i, worker in enumerate(self.workers):
            if isinstance(worker, RTTChanTCPWorker):
                if worker.port == port:
                    worker.close()
                    self.workers[i] = None



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from ..core import exceptions
from ..commands.base import CommandBase
from ..utility.rtt_server import RTTServer

LOG = logging.getLogger(__name__)

class ThreadsCommand(CommandBase):
    INFO = {
            'names': ['threads'],
            'group': 'gdbserver',
            'category': 'threads',
            'nargs': 1,
            'usage': "{flush,enable,disable,status}",
            'help': "Control thread awareness.",
            }

    def parse(self, args):
        self.action = args[0]
        if self.action not in ('flush', 'enable', 'disable', 'status'):
            raise exceptions.CommandError("invalid action")

    def execute(self):
        # Get the gdbserver for the selected core.
        core_number = self.context.selected_core.core_number
        try:
            gdbserver = self.context.session.gdbservers[core_number]
        except KeyError:
            raise exceptions.CommandError("no gdbserver for core #%i" % core_number)

        if gdbserver.thread_provider is None:
            self.context.write("Threads are unavailable")
            return

        if self.action == 'flush':
            gdbserver.thread_provider.invalidate()
            self.context.write("Threads flushed")
        elif self.action == 'enable':
            gdbserver.thread_provider.read_from_target = True
            self.context.write("Threads enabled")
        elif self.action == 'disable':
            gdbserver.thread_provider.read_from_target = False
            self.context.write("Threads disabled")
        elif self.action == 'status':
            self.context.write("Threads are " +
                    ("enabled" if gdbserver.thread_provider.read_from_target else "disabled"))

class ArmSemihostingCommand(CommandBase):
    INFO = {
            'names': ['arm'],
            'group': 'gdbserver',
            'category': 'semihosting',
            'nargs': 2,
            'usage': "semihosting {enable,disable}",
            'help': "Enable or disable semihosting.",
            'extra_help': "Provided for compatibility with OpenOCD. The same functionality can be achieved "
                            "by setting the 'enable_semihosting' session option.",
            }

    def parse(self, args):
        if args[0] != 'semihosting':
            raise exceptions.CommandError("invalid action")
        if args[1] not in ('enable', 'disable'):
            raise exceptions.CommandError("invalid action")
        self.action = args[1]

    def execute(self):
        enable = (self.action == 'enable')
        self.context.session.options['enable_semihosting'] = enable

class GdbserverMonitorInitCommand(CommandBase):
    """@brief 'init' command for OpenOCD compatibility.

    Many default gdbserver configurations send an 'init' monitor command.
    """
    INFO = {
            'names': ['init'],
            'group': 'gdbserver',
            'category': 'openocd_compatibility',
            'nargs': 2,
            'usage': "",
            'help': "Ignored; for OpenOCD compatibility.",
            }

    def execute(self):
        pass

class GdbserverMonitorExitCommand(CommandBase):
    """@brief 'exit' command to cleanly shut down the gdbserver from an IDE.

    This command is primarily intended to be used by an IDE to tell the pyocd process to exit when
    the debug session is terminated.
    """
    INFO = {
            'names': ['exit'],
            'group': 'gdbserver',
            'category': 'gdbserver',
            'nargs': 0,
            'usage': "",
            'help': "Terminate running gdbservers in this session.",
            'extra_help':
                "For the pyocd gdbserver subcommand, terminating gdbservers will cause the process to exit. The "
                "effect when the gdbserver(s) are running in a different environment depends on that program. "
                "Note that gdb will still believe the connection to be valid after this command completes, so "
                "executing the 'disconnect' command is a necessity."
            }

    def execute(self):
        for server in self.context.session.gdbservers.values():
            server.stop(wait=False)

class RTTCommand(CommandBase):
    INFO = {
            'names': ['rtt'],
            'group': 'gdbserver',
            'category': 'rtt',
            'nargs': "*",
            'usage': "rtt {setup,start,stop,channels,server}",
            'help': "Control SEGGER RTT compatible interface.",
            }

    def parse(self, args):
        if len(args) < 1:
            raise exceptions.CommandError("too few arguments")

        if args[0] == 'setup':
            if len(args) < 4:
                raise exceptions.CommandError("too few arguments")

            try:
                self.addr = int(args[1], 0)
                self.size = int(args[2], 0)
                self.id = " ".join(args[3:]).encode("utf-8")
            except ValueError as e:
                raise exceptions.CommandError("invalid action") from e
        elif args[0] == 'start' or args[0] == 'stop' or args[0] == 'channels':
            if len(args) > 1:
                raise exceptions.CommandError("too many arguments")
        elif args[0] == 'server':
            if len(args) < 2:
                    raise exceptions.CommandError("too few arguments")

            if args[1] == 'start':
                if len(args) < 4:
                    raise exceptions.CommandError("too few arguments")
                elif len(args) > 4:
                    raise exceptions.CommandError("too many arguments")

                try:
                    self.port = int(args[2], 0)
                    self.channel = int(args[3], 0)
                except ValueError as e:
                    raise exceptions.CommandError("invalid action*") from e
            elif args[1] == 'stop':
                if len(args) < 3:
                    raise exceptions.CommandError("too few arguments")
                elif len(args) > 3:
                    raise exceptions.CommandError("too many arguments")

                try:
                    self.port = int(args[2], 0)
                except ValueError as e:
                    raise exceptions.CommandError("invalid action") from e
            else:
                raise exceptions.CommandError("invalid action")

            self.server_action = args[1]
        else:
            raise exceptions.CommandError("invalid action")

        self.action = args[0]

    def execute(self):
        # Get the gdbserver for the selected core.
        core_number = self.context.selected_core.core_number
        try:
            gdbserver = self.context.session.gdbservers[core_number]
        except KeyError:
            raise exceptions.CommandError("no gdbserver for core #%i" % core_number)

        if self.action == "setup":
            if gdbserver.rtt_server is not None:
                if gdbserver.rtt_server.running:
                    raise exceptions.CommandError("rtt is already running")
                else:
                    gdbserver.rtt_server = None

            try:
                gdbserver.rtt_server = RTTServer(gdbserver.target, address = self.addr,
                                                 size = self.size,
                                                 control_block_id = self.id)
            except exceptions.RTTError as e:
                raise exceptions.CommandError(str(e)) from e
        elif self.action == "start":
            if gdbserver.rtt_server is None:
                raise exceptions.CommandError("rtt is not configured")

            try:
                gdbserver.rtt_server.start()
            except exceptions.RTTError as e:
                raise exceptions.CommandError(str(e)) from e

            self.context.write(f"Found RTT control block.")
        elif self.action == "stop":
            if gdbserver.rtt_server is not None:
                gdbserver.rtt_server.stop()
        elif self.action == "channels":
            control_block = gdbserver.rtt_server.control_block
            self.context.write(f"Channels: up={len(control_block.up_channels)}, "
                               f"down={len(control_block.down_channels)}")
            self.context.write("Up-channels:")
            for i, chan in enumerate(control_block.up_channels):
                name = chan.name if chan.name is not None else ""
                self.context.write(f"{i}: {name} {chan.size}")
            self.context.write("Down-channels:")
            for i, chan in enumerate(control_block.up_channels):
                name = chan.name if chan.name is not None else ""
                self.context.write(f"{i}: {name} {chan.size}")
        elif self.action == "server":
            if gdbserver.rtt_server is None:
                raise exceptions.CommandError("rtt is not configured")
            elif not gdbserver.rtt_server.running:
                raise exceptions.CommandError("rtt is not yet started")

            if self.server_action == "start":
                try:
                    gdbserver.rtt_server.add_server(self.port, self.channel)
                except exceptions.RTTError as e:
                    raise exceptions.CommandError(str(e)) from e
            elif self.server_action == "stop":
                try:
                    gdbserver.rtt_server.stop_server(self.port)
                except exceptions.RTTError as e:
                    raise exceptions.CommandError(str(e)) from e



# pyOCD debugger
# Copyright (c) 2020-2021 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import json
import threading
from typing import (Any, Optional, Tuple)

from .debug_probe import DebugProbe
from ..core import exceptions
from ..core.memory_interface import MemoryInterface
from ..core.plugin import Plugin
from ..utility.sockets import ClientSocket

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class TCPClientProbe(DebugProbe):
    """@brief Probe class that connects to a debug probe server.

    The protocol is a one-line JSON request and response form.

    Request structure:

    ````json
    {
        "id": <int>,
        "request": <str>,
        ["arguments": <list>]
    }
    ````

    Response structure:

    ````json
    {
        "id": <int>,
        "status": <int>,
        ["error": <str>,]
        ["result": <value>]
    }
    ````
    """

    DEFAULT_PORT = 5555

    PROTOCOL_VERSION = 1

    class StatusCode:
        """@brief Constants for errors reported from the server."""
        GENERAL_ERROR = 1
        PROBE_DISCONNECTED = 2
        PROBE_ERROR = 3
        TRANSFER_ERROR = 10
        TRANSFER_TIMEOUT = 11
        TRANSFER_FAULT = 12

    ## Map from status code to exception class.
    STATUS_CODE_CLASS_MAP = {
        StatusCode.GENERAL_ERROR: exceptions.Error,
        StatusCode.PROBE_DISCONNECTED: exceptions.ProbeDisconnected,
        StatusCode.PROBE_ERROR: exceptions.ProbeError,
        StatusCode.TRANSFER_ERROR: exceptions.TransferError,
        StatusCode.TRANSFER_TIMEOUT: exceptions.TransferTimeoutError,
        StatusCode.TRANSFER_FAULT: exceptions.TransferFaultError,
        }

    @classmethod
    def _extract_address(cls, unique_id):
        parts = unique_id.split(':', 1)
        if len(parts) == 1:
            port = cls.DEFAULT_PORT
        else:
            port = int(parts[1])
        return parts[0], port

    @classmethod
    def get_all_connected_probes(cls, unique_id=None, is_explicit=False):
        if is_explicit and unique_id is not None:
            return [cls(unique_id)]
        else:
            return []

    @classmethod
    def get_probe_with_id(cls, unique_id, is_explicit=False):
        return cls(unique_id) if is_explicit else None

    def __init__(self, unique_id):
        """@brief Constructor."""
        super(TCPClientProbe, self).__init__()
        hostname, port = self._extract_address(unique_id)
        self._uid = f"remote:{hostname}:{port}"
        self._socket = ClientSocket(hostname, port)
        self._is_open = False
        self._request_id = 0
        self._lock_count = 0
        self._lock_count_lock = threading.RLock()

    @property
    def vendor_name(self):
        return self._read_property('vendor_name', "vendor")

    @property
    def product_name(self):
        return self._read_property('product_name', "product")

    @property
    def supported_wire_protocols(self):
        return self._read_property('supported_wire_protocols')

    @property
    def unique_id(self):
        return self._uid

    @property
    def wire_protocol(self):
        return self._read_property('wire_protocol')

    @property
    def is_open(self):
        return self._is_open

    @property
    def capabilities(self):
        return self._read_property('capabilities')

    @property
    def request_id(self):
        """@brief Generate a new request ID."""
        rid = self._request_id
        self._request_id += 1
        return rid

    def _perform_request_without_raise(self, request: str, *args: Any) -> Tuple[Any, Optional[BaseException]]:
        """Execute a request-reply transaction with the server.

        The return value is a 2-tuple consisting of the optional result from the request and an optional
        exception object. The latter is only non-None if the request failed and a non-zero status code was
        returned.
        """
        # Protect requests with the local lock.
        with self._lock:
            rq = {
                    "id": self.request_id,
                    "request": request,
                }
            if len(args):
                rq["arguments"] = args
            formatted_request = json.dumps(rq)
            TRACE.debug("Request: %s", formatted_request)

            # Send request to server.
            self._socket.write(formatted_request.encode('utf-8') + b"\n")

            # Read response.
            response_data = self._socket.readline().decode('utf-8').strip()
            decoded_response = json.loads(response_data)
            TRACE.debug("decoded_response = %s", decoded_response)

            # Check for required keys.
            if ('id' not in decoded_response) or ('status' not in decoded_response):
                raise exceptions.ProbeError("malformed response from server; missing required field")

            # Check response status.
            exc = None
            status = decoded_response['status']
            if status != 0:
                # Get the error message.
                error = decoded_response.get('error', "(missing error message key)")
                LOG.debug("error received from server for command %s (status code %i): %s",
                        request, status, error)

                # Create an appropriate local exception based on the status code.
                exc = self.STATUS_CODE_CLASS_MAP.get(status, exceptions.ProbeError)(
                        "error received from server for command %s (status code %i): %s"
                        % (request, status, error))

            # Get response value. If not present then there was no return value from the command
            result = decoded_response.get('result', None)

            return result, exc

    def _perform_request(self, request: str, *args: Any) -> Any:
        """@brief Perform the request and immediately raise any errors."""
        result, exc = self._perform_request_without_raise(request, *args)
        if exc is not None:
            raise exc
        return result

    _PROPERTY_CONVERTERS = {
            'capabilities':                 lambda value: [DebugProbe.Capability[v] for v in value],
            'supported_wire_protocols':     lambda value: [DebugProbe.Protocol[v] for v in value],
            'wire_protocol':                lambda value: DebugProbe.Protocol[value] if (value is not None) else None,
        }

    def _read_property(self, name, default=None):
        if not self.is_open:
            return default
        result = self._perform_request('readprop', name)
        if name in self._PROPERTY_CONVERTERS:
            result = self._PROPERTY_CONVERTERS[name](result)
        return result

    def open(self):
        if not self._is_open:
            self._socket.connect()
            self._is_open = True
            self._socket.set_timeout(0.1)

        # Send hello message.
        self._perform_request('hello', self.PROTOCOL_VERSION)

        self._perform_request('open')

    def close(self):
        if self._is_open:
            self._perform_request('close')
            self._socket.close()
            self._is_open = False

    def lock(self):
        # The lock count is then used to only send the remote lock request once.
        with self._lock_count_lock:
            if self._lock_count == 0:
                self._perform_request('lock')
            self._lock_count += 1

    def unlock(self):
        # The remote unlock request is only sent when the outermost nested locking is unlocked.
        with self._lock_count_lock:
            assert self._lock_count > 0
            self._lock_count -= 1
            if self._lock_count == 0:
                self._perform_request('unlock')

    ## @name Target control
    ##@{

    def connect(self, protocol=None):
        if protocol is None:
            protocol = DebugProbe.Protocol.DEFAULT
        self._perform_request('connect', protocol.name)

    def disconnect(self):
        self._perform_request('disconnect')

    def swj_sequence(self, length, bits):
        self._perform_request('swj_sequence', length, bits)

    def swd_sequence(self, sequences):
        return self._perform_request('swd_sequence', sequences)

    def jtag_sequence(self, cycles, tms, read_tdo, tdi):
        return self._perform_request('jtag_sequence', cycles, tms, read_tdo, tdi)

    def set_clock(self, frequency):
        self._perform_request('set_clock', frequency)

    def reset(self):
        self._perform_request('reset')

    def assert_reset(self, asserted):
        self._perform_request('assert_reset', asserted)

    def is_reset_asserted(self):
        return self._perform_request('is_reset_asserted')

    def flush(self):
        self._perform_request('flush')

    ##@}

    ## @name DAP access
    ##@{

    def read_dp(self, addr, now=True):
        result, exc = self._perform_request_without_raise('read_dp', addr)

        def read_dp_cb():
            # Raise any exception here so the traceback includes the actual caller.
            if exc is not None:
                raise exc
            return result

        return read_dp_cb() if now else read_dp_cb

    def write_dp(self, addr, data):
        self._perform_request('write_dp', addr, data)

    def read_ap(self, addr, now=True):
        result, exc = self._perform_request_without_raise('read_ap', addr)

        def read_ap_cb():
            # Raise any exception here so the traceback includes the actual caller.
            if exc is not None:
                raise exc
            return result

        return read_ap_cb() if now else read_ap_cb

    def write_ap(self, addr, data):
        self._perform_request('write_ap', addr, data)

    def read_ap_multiple(self, addr, count=1, now=True):
        results, exc = self._perform_request_without_raise('read_ap_multiple', addr, count)

        def read_ap_multiple_cb():
            # Raise any exception here so the traceback includes the actual caller.
            if exc is not None:
                raise exc
            return results

        return read_ap_multiple_cb() if now else read_ap_multiple_cb

    def write_ap_multiple(self, addr, values):
        self._perform_request('write_ap_multiple', addr, values)

    def get_memory_interface_for_ap(self, ap_address):
        handle = self._perform_request('get_memory_interface_for_ap',
                ap_address.ap_version.value, ap_address.nominal_address)
        if handle is None:
            return None
        return RemoteMemoryInterface(self, handle)

    ##@}

    ## @name SWO
    ##@{

    def has_swo(self):
        return self._perform_request('has_swo')

    def swo_start(self, baudrate):
        self._perform_request('swo_start', baudrate)

    def swo_stop(self):
        self._perform_request('swo_stop')

    def swo_read(self):
        return self._perform_request('swo_read')

    ##@}

class RemoteMemoryInterface(MemoryInterface):
    """@brief Local proxy for a remote memory interface."""

    def __init__(self, remote_probe, handle):
        self._remote_probe = remote_probe
        self._handle = handle

    def write_memory(self, addr, data, transfer_size=32, **attrs):
        assert transfer_size in (8, 16, 32)
        self._remote_probe._perform_request('write_mem', self._handle, addr, data, transfer_size)

    def read_memory(self, addr, transfer_size=32, now=True, **attrs):
        assert transfer_size in (8, 16, 32)
        result, exc = self._remote_probe._perform_request_without_raise('read_mem', self._handle, addr, transfer_size)

        def read_callback():
            # Raise any exception here so the traceback includes the actual caller.
            if exc is not None:
                raise exc
            return result
        return read_callback() if now else read_callback

    def write_memory_block32(self, addr, data, **attrs):
        self._remote_probe._perform_request('write_block32', self._handle, addr, data)

    def read_memory_block32(self, addr, size, **attrs):
        return self._remote_probe._perform_request('read_block32', self._handle, addr, size)

    def write_memory_block8(self, addr, data, **attrs):
        self._remote_probe._perform_request('write_block8', self._handle, addr, data)

    def read_memory_block8(self, addr, size, **attrs):
        return self._remote_probe._perform_request('read_block8', self._handle, addr, size)

class TCPClientProbePlugin(Plugin):
    """@brief Plugin class for TCPClientProbePlugin."""

    def load(self):
        return TCPClientProbe

    @property
    def name(self):
        return "remote"

    @property
    def description(self):
        return "Client for the pyOCD debug probe server"




# pyOCD debugger
# Copyright (c) 2018-2019 Arm Limited
# Copyright (c) 2021 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from enum import Enum

from ..core.memory_map import MemoryType

LOG = logging.getLogger(__name__)

class FlashEraser(object):
    """@brief Class that manages high level flash erasing.

    Can erase a target in one of three modes:
    - chip erase: Erase all flash on the target.
    - mass erase: Also erase all flash on the target. However, on some targets, a mass erase has
        special properties such as unlocking security or erasing additional configuration regions
        that are not erased by a chip erase. If a target does not have a special mass erase, then
        it simply reverts to a chip erase.
    - sector erase: One or more sectors are erased.
    """
    class Mode(Enum):
        MASS = 1
        CHIP = 2
        SECTOR = 3

    def __init__(self, session, mode):
        """@brief Constructor.

        @param self
        @param session The session instance.
        @param mode One of the FlashEraser.Mode enums to select mass, chip, or sector erase.
        """
        self._session = session
        self._mode = mode

        ## Control used by the default Target.mass_erase() implementation to disable logging
        # inside ._chip_erase(), so the user doesn't see both mass- and chip-erase log messages
        # when chip-erase is used as the fallback.
        self._log_chip_erase = True

    def erase(self, addresses=None):
        """@brief Perform the type of erase operation selected when the object was created.

        For sector erase mode, an iterable of sector addresses specifications must be provided via
        the _addresses_ parameter. The address iterable elements can be either strings, tuples,
        or integers. Tuples must have two elements, the start and end addresses of a range to erase.
        Integers are simply an address within the single page to erase.

        String address specifications may be in one of three formats: "<address>", "<start>-<end>",
        or "<start>+<length>". Each field denoted by angled brackets is an integer literal in
        either decimal or hex notation.

        Examples:
        - "0x1000" - erase the one sector at 0x1000
        - "0x1000-0x4fff" - erase sectors from 0x1000 up to but not including 0x5000
        - "0x8000+0x800" - erase sectors starting at 0x8000 through 0x87ff

        @param self
        @param addresses List of addresses or address ranges of the sectors to erase.
        """
        if self._mode == self.Mode.MASS:
            self._mass_erase()
        elif self._mode == self.Mode.CHIP:
            self._chip_erase()
        elif self._mode == self.Mode.SECTOR and addresses:
            self._sector_erase(addresses)
        else:
            LOG.warning("No operation performed")

    def _mass_erase(self):
        LOG.info("Mass erasing device...")
        self._session.target.mass_erase()
        LOG.info("Mass erase complete")

    def _chip_erase(self):
        if self._log_chip_erase:
            LOG.info("Erasing chip...")
        # Erase all flash regions. This may be overkill if either each region's algo erases
        # all regions on the chip. But there's no current way to know whether this will happen,
        # so prefer to be certain.
        for region in self._session.target.memory_map.iter_matching_regions(type=MemoryType.FLASH):
            if region.flash is not None:
                if region.flash.is_erase_all_supported:
                    region.flash.init(region.flash.Operation.ERASE)
                    region.flash.erase_all()
                    region.flash.cleanup()
                else:
                    self._sector_erase([(region.start, region.end)])
        if self._log_chip_erase:
                LOG.info("Chip erase complete")

    def _sector_erase(self, addresses):
        flash = None
        currentRegion = None

        for spec in addresses:
            # Convert the spec into a start and end address.
            sector_addr, end_addr = self._convert_spec(spec)

            while sector_addr < end_addr:
                # Look up the flash memory region for the current address.
                region = self._session.target.memory_map.get_region_for_address(sector_addr)
                if region is None:
                    LOG.warning("address 0x%08x is not within a memory region", sector_addr)
                    break
                if not region.is_flash:
                    LOG.warning("address 0x%08x is not in flash", sector_addr)
                    break

                # Handle switching regions.
                if region is not currentRegion:
                    # Clean up previous flash.
                    if flash is not None:
                        flash.cleanup()

                    currentRegion = region
                    flash = region.flash
                    flash.init(flash.Operation.ERASE)

                assert flash is not None

                # Get sector info for the current address.
                sector_info = flash.get_sector_info(sector_addr)
                assert sector_info, ("sector address 0x%08x within flash region '%s' is invalid"
                                        % (sector_addr, region.name))

                # Align first page address.
                delta = sector_addr % sector_info.size
                if delta:
                    LOG.warning("sector address 0x%08x is unaligned", sector_addr)
                    sector_addr -= delta

                # Erase this page.
                LOG.info("Erasing sector 0x%08x (%d bytes)", sector_addr, sector_info.size)
                flash.erase_sector(sector_addr)

                sector_addr += sector_info.size

        if flash is not None:
            flash.cleanup()

    def _convert_spec(self, spec):
        if isinstance(spec, str):
            # Convert spec from string to range.
            if '-' in spec:
                a, b = spec.split('-')
                page_addr = int(a, base=0)
                end_addr = int(b, base=0)
            elif '+' in spec:
                a, b = spec.split('+')
                page_addr = int(a, base=0)
                length = int(b, base=0)
                end_addr = page_addr + length
            else:
                page_addr = int(spec, base=0)
                end_addr = page_addr + 1
        elif isinstance(spec, tuple):
            page_addr = spec[0]
            end_addr = spec[1]
        else:
            page_addr = spec
            end_addr = page_addr + 1
        return page_addr, end_addr





# pyOCD debugger
# Copyright (c) 2018-2020 Arm Limited
# Copyright (c) 2021-2023 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import errno
import itertools
import logging
import os
from typing import (IO, TYPE_CHECKING, Any, Callable, Dict, Iterator, List, Optional, Tuple, Union)

from elftools.elf.elffile import ELFFile
from intelhex import IntelHex

from ..core import exceptions
from .loader import (FlashLoader, ProgressCallback)

if TYPE_CHECKING:
    from ..core.session import Session

LOG = logging.getLogger(__name__)

def ranges(i: List[int]) -> Iterator[Tuple[int, int]]:
    """Accepts a sorted list of byte addresses. Breaks the addresses into contiguous ranges.
    Yields 2-tuples of the start and end address for each contiguous range.

    For instance, the input [0, 1, 2, 3, 32, 33, 34, 35] will yield the following 2-tuples:
    (0, 3) and (32, 35).
    """
    for a, b in itertools.groupby(enumerate(i), lambda x: x[1] - x[0]):
        b = list(b)
        yield b[0][1], b[-1][1]

class FileProgrammer(object):
    """@brief Class to manage programming a file in any supported format with many options.

    Most specifically, this class implements the behaviour provided by the command-line flash
    programming tool. The code in this class simply extracts data from the given file, potentially
    respecting format-specific options such as the base address for binary files. Then the heavy
    lifting of flash programming is handled by FlashLoader, and beneath that, FlashBuilder.

    Support file formats are:
    - Binary (.bin)
    - Intel Hex (.hex)
    - ELF (.elf or .axf)
    """
    def __init__(self,
            session: "Session",
            progress: Optional[ProgressCallback] = None,
            chip_erase: Optional[bool] = None,
            smart_flash: Optional[bool] = None,
            trust_crc: Optional[bool] = None,
            keep_unwritten: Optional[bool] = None,
            no_reset: Optional[bool] = None
        ):
        """@brief Constructor.

        @param self
        @param session The session object.
        @param progress A progress report handler as a callable that takes a percentage completed.
            If not set or None, a default progress handler will be used unless the session option
            'hide_programming_progress' is set to True, in which case progress will be disabled.
        @param chip_erase Sets whether to use chip erase or sector erase. The value must be one of
            "auto", "sector", or "chip". "auto" means the fastest erase method should be used.
        @param smart_flash If set to True, the programmer will attempt to not program pages whose
            contents are not going to change by scanning target flash memory. A value of False will
            force all pages to be erased and programmed.
        @param trust_crc Boolean indicating whether to use only the sector CRC32 to decide whether a
            sector already contains the data to be programmed. Use with caution, as CRC32 may return
            the same value for different content.
        @param keep_unwritten Depending on the sector versus page size and the amount of data
            written, there may be ranges of flash that would be erased but not written with new
            data. This parameter sets whether the existing contents of those unwritten ranges will
            be read from memory and restored while programming.
        @param no_reset Boolean indicating whether if the device should not be reset after the
            programming process has finished.
        """
        self._session = session
        self._chip_erase = chip_erase
        self._smart_flash = smart_flash
        self._trust_crc = trust_crc
        self._keep_unwritten = keep_unwritten
        self._no_reset = no_reset
        self._progress = progress
        self._loader = None

        self._format_handlers: Dict[str, Callable[..., None]] = {
            'axf': self._program_elf,
            'bin': self._program_bin,
            'elf': self._program_elf,
            'hex': self._program_hex,
            }

    def program(self, file_or_path: Union[str, IO[bytes]], file_format: Optional[str] = None, **kwargs: Any):
        """@brief Program a file into flash.

        @param self
        @param file_or_path Either a string that is a path to a file, or a file-like object.
        @param file_format Optional file format name, one of "bin", "hex", "elf", "axf". If not provided,
            the file's extension will be used. If a file object is passed for _file_or_path_ then
            this parameter must be used to set the format.
        @param kwargs Optional keyword arguments for format-specific parameters.

        The only current format-specific keyword parameters are for the binary format:
        - `base_address`: Memory address at which to program the binary data. If not set, the base
            of the boot memory will be used.
        - `skip`: Number of bytes to skip at the start of the binary file. Does not affect the
            base address.

        @exception FileNotFoundError Provided file_or_path string does not reference a file.
        @exception ValueError Invalid argument value, for instance providing a file object but
            not setting file_format.
        """
        is_path = isinstance(file_or_path, str)

        # Check for valid path first.
        if is_path and not os.path.isfile(file_or_path): # type: ignore # (type checker doesn't use is_path)
            raise FileNotFoundError(errno.ENOENT, "No such file: '{}'".format(file_or_path))

        # If no format provided, use the file's extension.
        if not file_format:
            if is_path:
                # Extract the extension from the path.
                file_format = os.path.splitext(file_or_path)[1][1:] # type: ignore # (type checker doesn't use is_path)

                # Explicitly check for no extension.
                if file_format == '':
                    raise ValueError("file path '{}' does not have an extension and "
                                        "no format is set".format(file_or_path))
            else:
                raise ValueError("file object provided but no format is set")

        # Check the format is one we understand.
        if file_format is None or file_format not in self._format_handlers:
            raise ValueError("unknown file format '%s'" % file_format)

        self._loader = FlashLoader(self._session,
                                    progress=self._progress,
                                    chip_erase=self._chip_erase,
                                    smart_flash=self._smart_flash,
                                    trust_crc=self._trust_crc,
                                    keep_unwritten=self._keep_unwritten,
                                    no_reset=self._no_reset)

        # file_obj = None
        # Open the file if a path was provided.
        if is_path:
            mode = 'rb'
            if file_format == 'hex':
                # hex file must be read as plain text file
                mode = 'r'
            assert isinstance(file_or_path, str)
            file_obj = open(file_or_path, mode)
        else:
            assert not isinstance(file_or_path, str)
            file_obj = file_or_path
        try:

            # Pass to the format-specific programmer.
            self._format_handlers[file_format](file_obj, **kwargs)
            self._loader.commit()
        finally:
            if is_path and file_obj is not None:
                file_obj.close()

    def _program_bin(self, file_obj: IO[bytes], **kwargs: Any) -> None:
        """@brief Binary file format loader"""
        assert self._loader

        # If no base address is specified use the start of the boot memory.
        address = kwargs.get('base_address', None)
        if address is None:
            assert self._session.target
            boot_memory = self._session.target.memory_map.get_boot_memory()
            if boot_memory is None:
                raise exceptions.TargetSupportError("No boot memory is defined for this device")
            address = boot_memory.start
        assert isinstance(address, int)

        skip_offset = kwargs.get('skip', 0)
        if not isinstance(skip_offset, int):
            raise TypeError("skip argument must be an integer")
        file_obj.seek(skip_offset, os.SEEK_SET)
        data = list(bytearray(file_obj.read()))

        self._loader.add_data(address, data)

    def _program_hex(self, file_obj: IO[bytes], **kwargs: Any) -> None:
        """Intel hex file format loader"""
        assert self._loader

        hexfile = IntelHex(file_obj)
        addresses = hexfile.addresses()
        addresses.sort()

        data_list = list(ranges(addresses))
        for start, end in data_list:
            size = end - start + 1
            data = list(hexfile.tobinarray(start=start, size=size))
            # Ignore invalid addresses for HEX files only
            # Binary files (obviously) don't contain addresses
            # For ELF files, any metadata that's not part of the application code
            # will be held in a section that doesn't have the SHF_WRITE flag set
            try:
                self._loader.add_data(start, data)
            except ValueError as e:
                LOG.warning("Failed to add data chunk: %s", e)

    def _program_elf(self, file_obj: IO[bytes], **kwargs: Any) -> None:
        assert self._loader

        elf = ELFFile(file_obj)
        for segment in elf.iter_segments():
            addr = segment['p_paddr']
            if segment.header.p_type == 'PT_LOAD' and segment.header.p_filesz != 0:
                data = bytearray(segment.data())
                LOG.debug("Writing segment LMA:0x%08x, VMA:0x%08x, size %d", addr,
                          segment['p_vaddr'], segment.header.p_filesz)
                try:
                    self._loader.add_data(addr, data)
                except ValueError as e:
                    LOG.warning("Failed to add data chunk: %s", e)
            else:
                LOG.debug("Skipping segment LMA:0x%08x, VMA:0x%08x, size %d", addr,
                          segment['p_vaddr'], segment.header.p_filesz)



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import logging
from typing import (Dict, Iterable, Optional, Set)

LOG = logging.getLogger(__name__)
TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.CRITICAL)

class Scope:
    """@brief Debug sequence execution scope.

    Scopes have both a link to a parent scope. The former is used to read regular variables defined in
    super-scopes. Writing a variable always sets it in the scope to in which it was defined, unless the
    variable hasn't been set before in which case it is set in the scope that was called.
    """

    def __init__(
                self,
                parent: Optional["Scope"] = None,
                name: str = ""
            ) -> None:
        """@brief Constructor.
        @param self The Scope object.
        @param parent Optional parent scope reference. If not provided or set to None, the new scope
            becomes a root.
        @param specials Optional specials scope reference. If not provided or set to None, and a
            parent was provided, then the specials scope from the parent is used.
        @param name Optional name for the scope.
        """
        self._name = name
        self._parent = parent
        self._variables: Dict[str, int] = {} # Map from name: value.
        # A variable is read-only if its name is in this set. Start off with
        self._ro_variables: Set[str] = set()

    @property
    def name(self) -> str:
        """@brief The scope's name."""
        return self._name

    @property
    def parent(self) -> Optional[Scope]:
        """@brief Parent scope.

        The parent of the root scope is None.
        """
        return self._parent

    @property
    def variables(self) -> Set[str]:
        """@brief Set of the names of all variables defined in this scope.

        Does not include any variables from parent scopes.
        """
        return set(self._variables.keys())

    def get(self, name: str) -> int:
        """@brief Read a variable."""
        try:
            value = self._variables[name]
        except KeyError:
            if self._parent is not None:
                value = self._parent.get(name)
            else:
                raise
        TRACE.debug("get '%s' -> 0x%016x", name, value)
        return value

    def set(self, name: str, value: int, readonly: bool = False) -> None:
        """@brief Write a variable.

        The variable is set in the scope in which it is defined (following the parent links). The first
        time a variable is set, it can be marked as read-only via the `readonly` parameter.

        @param self
        @param name Name of the variable.
        @param value Integer value of the variable. Limited to 64-bit.
        @param readonly If the variable has not been previously defined, this parameter determines if it
            is writable. Primarily used for debugvars and other predefined variables.
        """
        TRACE.debug("set '%s' <- 0x%016x", name, value)

        # Catch attempt to rewrite a read-only variable.
        if self.is_read_only(name):
            raise RuntimeError("attempt to modify read-only variable '%s'" % name)

        scope = self
        while scope is not None:
            if name in scope._variables:
                # Found a scope with the variable definition.
                scope._variables[name] = value
                return
            scope = scope.parent

        # An existing definition of the variable wasn't found, so add it to the this scope.
        self._variables[name] = value
        if readonly:
            self._ro_variables.add(name)

    def copy_variables(self, from_scope: Scope, variables: Iterable[str]) -> None:
        """@brief Copy a set of variables from another scope into this one.

        @param self
        @param from_scope Scope to copy from. This scope must not be connected via parent links to the
            called scope, or the behaviour is undefined.
        @param variables Iterable of variable names which will be copied. If a variable is not defined
            in the `from_scope`, it will not be copied.
        """
        for name in variables:
            if from_scope.is_defined(name):
                self.set(name, from_scope.get(name))

    def is_defined(self, name: str, recurse_parents: bool = True) -> bool:
        """@brief Returns whether a variable has been set in this or any linked scope.

        @param self
        @param name Name of the variable to query.
        @param recurse_parents Whether parent scopes should also be queried if this scope doesn't contain the specified
            variable.
        @return Boolean of whether the named variable has been defined.
        """
        if name in self._variables:
            return True
        elif recurse_parents and (self._parent is not None):
            return self._parent.is_defined(name, recurse_parents)
        else:
            return False

    def freeze(self) -> None:
        """@brief Make all variables defined in the scope read-only."""
        self._ro_variables = set(self._variables.keys())

    def is_read_only(self, name: str) -> bool:
        """@brief Returns a boolean for whether the named variable is read-only.

        First checks the called scope. If the variable isn't found to be read-only, it asks the
        parent. Thus, once a variable is marked as read-only, it remains read-only in all child
        scopes.
        """
        if name in self._ro_variables:
            return True
        elif self.parent is not None:
            return self.parent.is_read_only(name)
        else:
            return False

    def _build_dump(self, indent: str) -> str:
        """@brief Construct a scope dump with a given ident level."""
        s = f"<{type(self).__name__}@{id(self):x} {self.name}\n"
        if self.parent:
            parent_str = self.parent._build_dump(indent + '  ')
        else:
            parent_str = "None"
        s += f"{indent}parent={parent_str}\n"
        s += f"{indent}variables=[\n"
        for n, v in self._variables.items():
            s += f"{indent}  '{n}'={v:#x} {'(RO)' if n in self._ro_variables else '(RW)'}\n"
        s += f"{indent}]>"
        return s

    def dump(self) -> str:
        """@brief Return a string detailing the scope, its parents, and all variable values."""
        return self._build_dump(indent="  ")

    def __len__(self) -> int:
        """@brief Return the number of variables in this scope (not including parents)."""
        return len(self._variables)

    def __repr__(self) -> str:
        """@brief Shortened representation of the scope without variable values."""
        return f"<{type(self).__name__}@{id(self):x} {self.name} "\
               f"parent={self._parent!r} [{', '.join(self._variables.keys())}]>"



# pyOCD debugger
# Copyright (c) 2020 Arm Limited
# Copyright (c) 2021-2022 Chris Reed
# Copyright (c) 2023 Marian Muller Rebeyrol
# SPDX-License-Identifier: Apache-2.0
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import six
import logging
from time import sleep
import pylink
from pylink.enums import JLinkInterfaces
from pylink.errors import (JLinkException, JLinkWriteException, JLinkReadException)
from typing import (TYPE_CHECKING, Optional, Tuple, Any, Sequence, Union, Callable)

from .debug_probe import DebugProbe
from ..core.memory_interface import MemoryInterface
from ..core import exceptions
from ..core.plugin import Plugin
from ..core.options import OptionInfo
from ..utility import conversion

if TYPE_CHECKING:
    from pylink.structs import JLinkHardwareStatus

LOG = logging.getLogger(__name__)

TRACE = LOG.getChild("trace")
TRACE.setLevel(logging.WARNING)

## @brief Wraps a JLink as a DebugProbe.
class JLinkProbe(DebugProbe):

    # Address of DP's SELECT register.
    DP_SELECT = 0x8

    # Bitmasks for AP register address fields.
    A32 = 0x0000000c
    APBANKSEL = 0x000000f0
    APSEL = 0xff000000
    APSEL_APBANKSEL = APSEL | APBANKSEL

    @classmethod
    def _get_jlink(cls) -> Optional[pylink.JLink]:
        # TypeError is raised by pylink if the JLink DLL cannot be found.
        try:
            return pylink.JLink(
                    log=TRACE.info,
                    detailed_log=TRACE.debug,
                    error=TRACE.error,
                    warn=TRACE.warn,
                    )
        except TypeError:
            return None

    @classmethod
    def _format_serial_number(cls, serial_number):
        return "{:d}".format(serial_number)

    @classmethod
    def get_all_connected_probes(cls, unique_id=None, is_explicit=False):
        try:
            jlink = cls._get_jlink()
            if jlink is None:
                return []
            return [cls(cls._format_serial_number(info.SerialNumber)) for info in jlink.connected_emulators()]
        except JLinkException as exc:
            raise cls._convert_exception(exc) from exc

    @classmethod
    def get_probe_with_id(cls, unique_id, is_explicit=False):
        try:
            jlink = cls._get_jlink()
            if jlink is None:
                return None
            for info in jlink.connected_emulators():
                sn = cls._format_serial_number(info.SerialNumber)
                if sn == unique_id:
                    return cls(sn)
            return None
        except JLinkException as exc:
            raise cls._convert_exception(exc) from exc

    @classmethod
    def _get_probe_info(cls, serial_number, jlink):
        """@brief Look up and return a JLinkConnectInfo for the probe with matching serial number.
        @param cls The class object.
        @param serial_number String serial number. Must be the full serial number.
        @return JLinkConnectInfo object or None if there was no match.
        """
        try:
            for info in jlink.connected_emulators():
                if cls._format_serial_number(info.SerialNumber) == serial_number:
                    return info
            return None
        except JLinkException as exc:
            raise cls._convert_exception(exc) from exc

    def __init__(self, serial_number):
        """@brief Constructor.
        @param self The object.
        @param serial_number String. The J-Link's serial number.
        """
        super().__init__()
        link = self._get_jlink()
        if link is None:
            raise exceptions.ProbeError("unable to open JLink DLL")
        self._link = link

        info = self._get_probe_info(serial_number, self._link)
        if info is None:
            raise exceptions.ProbeError("could not find JLink probe with serial number '{}'".format(serial_number))

        self._serial_number = serial_number
        self._serial_number_int = int(serial_number, base=10)
        self._supported_protocols = None
        self._protocol = None
        self._default_protocol = None
        self._is_open = False
        self._product_name = six.ensure_str(info.acProduct)
        self._memory_interfaces = {}

    @property
    def description(self):
        return self.vendor_name + " " + self.product_name

    @property
    def vendor_name(self):
        return "Segger"

    @property
    def product_name(self):
        return self._product_name

    ## @brief Only valid after opening.
    @property
    def supported_wire_protocols(self):
        return self._supported_protocols

    @property
    def unique_id(self):
        return self._serial_number

    @property
    def wire_protocol(self):
        return self._protocol

    @property
    def is_open(self):
        return self._link.opened

    @property
    def capabilities(self):
        return {
                self.Capability.SWO,
                self.Capability.BANKED_DP_REGISTERS,
                self.Capability.APv2_ADDRESSES,
                self.Capability.PIN_ACCESS,
                }

    def get_accessible_pins(self, group: DebugProbe.PinGroup) -> Tuple[int, int]:
        """@brief Return masks of pins accessible via the .read_pins()/.write_pins() methods.

        @return Tuple of pin masks for (0) readable, (1) writable pins. See DebugProbe.Pin for mask
        values for those pins that have constants.
        """
        if group is DebugProbe.PinGroup.PROTOCOL_PINS:
            return (self.ProtocolPin.ALL_PINS, self.ProtocolPin.ALL_PINS)
        else:
            return (0, 0)

    def open(self):
        assert self.session

        try:
            # Configure UI usage. We must do this here rather than in the ctor because the ctor
            # doesn't have access to the session.
            if self.session.options.get('jlink.non_interactive'):
                self._link.disable_dialog_boxes()

            self._link.open(self._serial_number_int)
            self._is_open = True

            # Get available wire protocols.
            ifaces = self._link.supported_tifs()
            self._supported_protocols = [DebugProbe.Protocol.DEFAULT]
            if ifaces & (1 << JLinkInterfaces.JTAG):
                self._supported_protocols.append(DebugProbe.Protocol.JTAG)
            if ifaces & (1 << JLinkInterfaces.SWD):
                self._supported_protocols.append(DebugProbe.Protocol.SWD)
            if not len(self._supported_protocols) >= 2: # default + 1
                raise exceptions.ProbeError("J-Link probe {} does not support any known wire protocols".format(
                        self.unique_id))

            # Select default protocol, preferring SWD over JTAG.
            if DebugProbe.Protocol.SWD in self._supported_protocols:
                self._default_protocol = DebugProbe.Protocol.SWD
            else:
                self._default_protocol = DebugProbe.Protocol.JTAG
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def close(self):
        try:
            self._link.close()
            self._is_open = False
            self._memory_interfaces = {}
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    # ------------------------------------------- #
    #          Target control functions
    # ------------------------------------------- #
    def connect(self, protocol=None):
        """@brief Connect to the target via JTAG or SWD."""
        assert self.session

        # Handle default protocol.
        if (protocol is None) or (protocol == DebugProbe.Protocol.DEFAULT):
            protocol = self._default_protocol

        # Validate selected protocol.
        assert self._supported_protocols is not None
        if protocol not in self._supported_protocols:
            raise ValueError("unsupported wire protocol %s" % protocol)

        # Convert protocol to port enum.
        if protocol == DebugProbe.Protocol.SWD:
            iface = JLinkInterfaces.SWD
        elif protocol == DebugProbe.Protocol.JTAG:
            iface = JLinkInterfaces.JTAG
        else:
            raise exceptions.InternalError(f"unknown wire protocol ({protocol})")

        try:
            self._link.set_tif(iface)
            if self.session.options.get('jlink.power'):
                self._link.power_on()

            # Connect if a device name was supplied.
            device_name = self.session.options.get('jlink.device')
            if device_name is not None:
                self._link.connect(device_name)

            self._link.coresight_configure()
            self._protocol = protocol
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def swj_sequence(self, length, bits):
        for chunk in range((length + 31) // 32):
            chunk_word = bits & 0xffffffff
            chunk_len = min(length, 32)

            if chunk_len == 32:
                self._link.swd_write32(chunk_len, chunk_word)
            else:
                self._link.swd_write(0, chunk_word, chunk_len)

            bits >>= 32
            length -= 32

        self._link.swd_sync()

    def disconnect(self):
        """@brief Disconnect from the target."""
        assert self.session
        try:
            if self.session.options.get('jlink.power'):
                self._link.power_off()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

        self._protocol = None

    def set_clock(self, frequency):
        try:
            self._link.set_speed(int(frequency) // 1000)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def reset(self):
        assert self.session
        try:
            self._link.set_reset_pin_low()
            sleep(self.session.options.get('reset.hold_time'))
            self._link.set_reset_pin_high()
            sleep(self.session.options.get('reset.post_delay'))
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def assert_reset(self, asserted):
        try:
            if asserted:
                self._link.set_reset_pin_low()
            else:
                self._link.set_reset_pin_high()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def is_reset_asserted(self):
        try:
            status = self._link.hardware_status
            return status.tres == 0
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def read_pins(self, group: DebugProbe.PinGroup, mask: int) -> int:
        """@brief Read values of selected debug probe pins.

        See DebugProbe.ProtocolPin for mask values.

        @param self
        @param group Select the pin group to read.
        @param mask Bit mask indicating which pins will be read. The return value will contain only
            bits set in this mask.
        @return Bit mask with the current value of selected pins at each pin's relevant bit position.
       """
        try:
            if group is DebugProbe.PinGroup.PROTOCOL_PINS:
                status = self._link.hardware_status
                return self.from_jlink_pins(status) & mask
            else:
                return 0
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def write_pins(self, group: DebugProbe.PinGroup, mask: int, value: int) -> None:
        """@brief Set values of selected debug probe pins.

        See DebugProbe.ProtocolPin for mask values.

        @param self
        @param group Select the pin group to read.
        @param mask Bit mask indicating which pins will be written.
        @param value Mask containing the bit value of to written for selected pins at each pin's
            relevant bit position..
        """
        assert self._link
        try:
            if group is not DebugProbe.PinGroup.PROTOCOL_PINS:
                return
            if mask & DebugProbe.ProtocolPin.SWCLK_TCK:
                if value & DebugProbe.ProtocolPin.SWCLK_TCK:
                    self._link.set_tck_pin_high()
                else:
                    self._link.set_tck_pin_low()
            if mask & DebugProbe.ProtocolPin.SWDIO_TMS:
                if value & DebugProbe.ProtocolPin.SWDIO_TMS:
                    self._link.set_tms_pin_high()
                else:
                    self._link.set_tms_pin_low()
            if mask & DebugProbe.ProtocolPin.TDI:
                if value & DebugProbe.ProtocolPin.TDI:
                    self._link.set_tdi_pin_high()
                else:
                    self._link.set_tdi_pin_low()
            if mask & DebugProbe.ProtocolPin.nRESET:
                if value & DebugProbe.ProtocolPin.nRESET:
                    self._link.set_reset_pin_high()
                else:
                    self._link.set_reset_pin_low()
            if mask & DebugProbe.ProtocolPin.nTRST:
                if value & DebugProbe.ProtocolPin.nTRST:
                    self._link.set_trst_pin_high()
                else:
                    self._link.set_trst_pin_low()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    @staticmethod
    def from_jlink_pins(status: "JLinkHardwareStatus") -> int:
        # JLinkHardwareStatus attributes:
        # - tck: measured state of TCK pin.
        # - tdi: measured state of TDI pin.
        # - tdo: measured state of TDO pin.
        # - tms: measured state of TMS pin.
        # - tres: measured state of TRES pin.
        # - trst: measured state of TRST pin.
        result = 0
        if status.tck:
            result |= DebugProbe.ProtocolPin.SWCLK_TCK
        if status.tms:
            result |= DebugProbe.ProtocolPin.SWDIO_TMS
        if status.tdi:
            result |= DebugProbe.ProtocolPin.TDI
        if status.tdo:
            result |= DebugProbe.ProtocolPin.TDO
        if status.tres:
            result |= DebugProbe.ProtocolPin.nRESET
        if status.trst:
            result |= DebugProbe.ProtocolPin.nTRST
        return result

    # ------------------------------------------- #
    #          DAP Access functions
    # ------------------------------------------- #

    def read_dp(self, addr, now=True):
        try:
            value = self._link.coresight_read(addr // 4, ap=False)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc
        else:
            def read_reg_cb():
                return value

            return value if now else read_reg_cb

    def write_dp(self, addr, data):
        try:
            self._link.coresight_write(addr // 4, data, ap=False)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def read_ap(self, addr, now=True):
        assert isinstance(addr, int)
        try:
            value = self._link.coresight_read((addr & self.A32) // 4, ap=True)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc
        else:
            def read_reg_cb():
                return value

            return value if now else read_reg_cb

    def write_ap(self, addr, data):
        assert isinstance(addr, int)
        try:
            self._link.coresight_write((addr & self.A32) // 4, data, ap=True)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def read_ap_multiple(self, addr, count=1, now=True):
        results = [self.read_ap(addr, now=True) for n in range(count)]

        def read_ap_multiple_result_callback():
            return results

        return results if now else read_ap_multiple_result_callback

    def write_ap_multiple(self, addr, values):
        for v in values:
            self.write_ap(addr, v)

    def get_memory_interface_for_ap(self, ap_address):
        assert self._is_open
        # JLink memory access commands only support AP 0
        if ap_address.apsel != 0:
            return None
        # JLink memory access commands require to be conneected to the target
        if not self._link.target_connected():
            return None
        apsel = ap_address.apsel
        if apsel not in self._memory_interfaces:
            self._memory_interfaces[apsel] = JLinkMemoryInterface(self._link, apsel)
        return self._memory_interfaces[apsel]

    def swo_start(self, baudrate):
        try:
            self._link.swo_start(int(baudrate))
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def swo_stop(self):
        try:
            self._link.swo_stop()
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    def swo_read(self):
        try:
            return self._link.swo_read(0, self._link.swo_num_bytes(), True)
        except JLinkException as exc:
            raise self._convert_exception(exc) from exc

    @staticmethod
    def _convert_exception(exc):
        if isinstance(exc, JLinkException):
            # J-Link returns this unhelpful error when it's really a transfer fault. The JLinkWriteException
            # and JLinkReadException exceptions checked below seem to only be returned for the higher level
            # read/write APIs.
            if str(exc) == "Unspecified error.":
                return exceptions.TransferFaultError(str(exc))
            else:
                return exceptions.ProbeError(str(exc))
        elif isinstance(exc, (JLinkWriteException, JLinkReadException)):
            return exceptions.TransferFaultError(str(exc))
        else:
            return exc

class JLinkMemoryInterface(MemoryInterface):
    """@brief Concrete memory interface for a single AP."""

    def __init__(self, link, apsel):
        self._link = link
        self._apsel = apsel

    def write_memory(self, addr: int, data: int, transfer_size: int=32, **attrs: Any) -> None:
        """@brief Write a single memory location.

        By default the transfer size is a word.
        """
        assert transfer_size in (8, 16, 32)
        addr &= 0xffffffff
        if transfer_size == 32:
            self._link.memory_write32(addr, [data])
        elif transfer_size == 16:
            self._link.memory_write16(addr, [data])
        elif transfer_size == 8:
            self._link.memory_write8(addr, [data])

    def read_memory(self, addr: int, transfer_size: int=32, now: bool=True, **attrs: Any) \
            -> Union[int, Callable[[], int]]:
        """@brief Read a memory location.

        By default, a word will be read.
        """
        assert transfer_size in (8, 16, 32)
        addr &= 0xffffffff
        if transfer_size == 32:
            result = self._link.memory_read32(addr, 1)[0]
        elif transfer_size == 16:
            result = self._link.memory_read16(addr, 1)[0]
        elif transfer_size == 8:
            result = self._link.memory_read8(addr, 1)[0]

        def read_callback():
            return result
        return result if now else read_callback

    def write_memory_block32(self, addr: int, data: Sequence[int], **attrs: Any) -> None:
        addr &= 0xffffffff
        self._link.memory_write32(addr, data)

    def read_memory_block32(self, addr: int, size: int, **attrs: Any) -> Sequence[int]:
        addr &= 0xffffffff
        return self._link.memory_read32(addr, size)

    def read_memory_block8(self, addr: int, size: int, **attrs: Any) -> Sequence[int]:
        addr &= 0xffffffff
        res = []

        # Transfers are handled in 3 phases:
        #   1. read 8-bit chunks until the first aligned address is reached,
        #   2. read 32-bit chunks from all aligned addresses,
        #   3. read 8-bit chunks from the remaining unaligned addresses.
        # If the requested size is so small that phase-1 would not even reach
        # aligned address, go straight to phase-3.

        # 1. read leading unaligned bytes
        unaligned_count = 3 & (4 - addr)
        if (size > unaligned_count > 0):
            res += self._link.memory_read8(addr, unaligned_count)
            size -= unaligned_count
            addr += unaligned_count

        # 2. read aligned block of 32 bits
        if (size >= 4):
            aligned_size = size & ~3
            res += conversion.u32le_list_to_byte_list(self._link.memory_read32(addr, aligned_size//4))
            size -= aligned_size
            addr += aligned_size

        # 3. read trailing unaligned bytes
        if (size > 0):
            res += self._link.memory_read8(addr, size)

        return res

    def write_memory_block8(self, addr: int, data: Sequence[int], **attrs: Any) -> None:
        addr &= 0xffffffff
        size = len(data)
        idx = 0

        # write leading unaligned bytes
        unaligned_count = 3 & (4 - addr)
        if (size > unaligned_count > 0):
            self._link.memory_write8(addr, data[:unaligned_count])
            size -= unaligned_count
            addr += unaligned_count
            idx += unaligned_count

        # write aligned block of 32 bits
        if (size >= 4):
            aligned_size = size & ~3
            self._link.memory_write32(addr, conversion.byte_list_to_u32le_list(data[idx:idx + aligned_size]))
            size -= aligned_size
            addr += aligned_size
            idx += aligned_size

        # write trailing unaligned bytes
        if (size > 0):
            self._link.memory_write8(addr, data[idx:])

class JLinkProbePlugin(Plugin):
    """@brief Plugin class for JLinkProbe."""

    def should_load(self):
        """@brief Load the J-Link plugin if the J-Link library is available."""
        return JLinkProbe._get_jlink() is not None

    def load(self):
        return JLinkProbe

    @property
    def name(self):
        return "jlink"

    @property
    def description(self):
        return "SEGGER J-Link debug probe"

    @property
    def options(self):
        """@brief Returns J-Link probe options."""
        return [
            OptionInfo('jlink.device', str, None,
                "If this option is set to a supported J-Link device name, then the J-Link will be asked connect "
                "using this name. Otherwise, the J-Link is configured for only the low-level CoreSight operations "
                "required by pyOCD. Ordinarily, it does not need to be set."),
            OptionInfo('jlink.power', bool, True,
                "Enable target power when connecting via a JLink probe, and disable power when "
                "disconnecting. Default is True."),
            OptionInfo('jlink.non_interactive', bool, True,
                "Controls whether the J-Link DLL is allowed to present UI dialog boxes and its control "
                "panel. Note that dialog boxes will actually still be visible, but the default option "
                "will be chosen automatically after 5 seconds. Default is True."),
            ]
